/*
 *  Copyright (C) 1998-99 Luca Deri <deri@unipi.it>
 *                        Stefano Suin <stefano@unipi.it>
 *                        (Stefano contributed to version 1.0)
 *                      
 *		  	  Centro SERRA, University of Pisa
 *		 	  http://www-serra.unipi.it/
 *  					
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
/*
 * Copyright (c) 1994, 1996
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "ntop.h"

#define DONT_USE_FORK 1
/* #define DEBUG */


/* Extern */
extern void init_counters();
extern void cleanup_curses();
extern int  checkKeyPressed();
extern void returnHTTPaccessDenied();
extern int  checkHTTPpassword(char *pw);
extern void readConfigurationFile();
extern void printTCPSessions();
extern void printUDPSessions();
extern void printIPpkt();
extern void printTCPpkt();
extern void handleHTTPrequest();
extern void clrscr();
extern void error(const char *fmt, ...);
extern void scanTimedoutTCPSessions();
extern void purgeIdleHosts();
extern void initIPServices();
extern RETSIGTYPE _printHostsTraffic(int signumber_ignored);
extern RETSIGTYPE printHostsTraffic(int signumber_ignored, int);
extern time_t nextSessionTimeoutScan;
extern char *optarg, version[];
extern ServiceEntry *udpSvc[SERVICE_HASH_SIZE], *tcpSvc[SERVICE_HASH_SIZE];
extern void updateThpt();
extern void LogStatsToFile();
extern void handleLocalAddresses(char* addresses);
extern void handleFlowsSpecs(char* flows);
extern void init_curses();
#ifdef MULTITHREADED
extern void* dequeuePacket(void*);
#ifdef ASYNC_ADDRESS_RESOLUTION
extern void* dequeueAddress(void*);
#endif
extern void queuePacket(u_char *unused, const struct pcap_pkthdr *h,
			const u_char *p);
extern void* handleCursesRefresh(void* notUsed);
#endif
#ifdef HAVE_LSOF 
extern void read_lsofInfo(int);
#endif
extern void getLocalHostAddress(struct in_addr *hostAddress, char* device);


/* Globals */
char *program_name, *device;
time_t actTime, initialSniffTime, lastRefreshTime;
int datalink, logFlag=0, showProtoWarnings = 0;
#ifdef DEBUG
int numChildren;
#endif
short printingUsage;
pcap_t *pcapPtr;
char *protoIPTrafficInfos[MAX_NUM_HANDLED_IP_PROTOCOLS]; /* array 0-numIpProtosToMonitor */
u_short numIpProtosToMonitor=0, numIpPortsToHandle=0;
int ipPortMapper[TOP_IP_PORT];
char* rFileName=NULL;
#ifdef MULTITHREADED
pthread_mutex_t     packetQueueMutex, hostsHashMutex; 
#ifdef HAVE_LSOF
pthread_mutex_t     lsofMutex;
#endif
pthread_t           dequeueThreadId, handleWebConnectionsThreadId, 
                    cursesRefreshThreadId, thptUpdateThreadId, 
                    scanIdleThreadId, logFileLoopThreadId,
                    lsofThreadId;
PacketInformation packetQueue[PACKET_QUEUE_LENGTH];
#ifdef ASYNC_ADDRESS_RESOLUTION
struct hnamemem   *addressQueue[ADDRESS_QUEUE_LENGTH];
#endif
unsigned int packetQueueLen, maxPacketQueueLen, packetQueueHead, packetQueueTail;
TrafficCounter droppedPackets;
#ifdef USE_SEMAPHORES
sem_t queueSem;
#ifdef ASYNC_ADDRESS_RESOLUTION
sem_t queueAddressSem;
#endif
#else
ConditionalVariable queueCondvar;
#ifdef ASYNC_ADDRESS_RESOLUTION
ConditionalVariable queueAddressCondvar;
#endif
#endif
#ifdef ASYNC_ADDRESS_RESOLUTION
pthread_t dequeueAddressThreadId;
unsigned int addressQueueLen, maxAddressQueueLen, addressQueueHead, addressQueueTail;
TrafficCounter droppedAddresses;
pthread_mutex_t addressQueueMutex;
#endif
#endif

/* Locals */
static int sock;

/* Forwards */
RETSIGTYPE cleanup(int);
RETSIGTYPE handleDiedChild(int);
RETSIGTYPE dontFreeze(int);
void detachFromTerminal();
void usage(int);
void handleProtocols(char*);
void addDefaultProtocols();
short handleProtocol(char* protoName, char *protocol);
void handleProtocolList(char* protoName, char *protocolList);
void* handleWebConnections(void*);
#ifdef MULTITHREADED
void* updateThptLoop(void* notUsed);
void* scanIdleLoop(void* notUsed);
void* logFileLoop(void* notUsed);
void* periodicLsofLoop(void* notUsed);
#endif

/* Global */
unsigned int localnet, netmask;
struct in_addr localHostAddress;
int numericFlag, refreshRate, localAddrFlag, percentMode;
int idleFlag, webMode, newSock, logTimeout;
time_t nextLogTime=0;
FILE *logd;
FlowFilterList *flowsList;

/* *************************** */

void ignoreSignal(int signalId) {
  close(newSock);
#ifndef DONT_USE_FORK
  exit(0);
#endif
}
 
/* *************************** */

int main(int argc, char *argv[])
{
  int op, pflag, rc, webPort;
  char *cp, *expression, *localAddresses=NULL;
  RETSIGTYPE (*oldhandler)(int); 
  char ebuf[PCAP_ERRBUF_SIZE], flowSpecs[2048];
  struct sockaddr_in sin;
  time_t lastTime;
  struct bpf_program fcode;

  flowSpecs[0] = '\0';
  flowsList = NULL;
  localAddrFlag = 1;
  logTimeout = 0;
  device = NULL;
  pflag = 0;
  numericFlag=0;
  percentMode = 0;
  refreshRate = ALARM_TIME; 
  idleFlag = 1;
  webMode = 0;
  logd = NULL;
  printingUsage = 0;

  actTime = time(NULL);

  if ((cp = strrchr(argv[0], '/')) != NULL)
    program_name = cp + 1;
  else
    program_name = argv[0];

  memset(protoIPTrafficInfos, 0, sizeof(protoIPTrafficInfos));
  memset(ipPortMapper, -1, sizeof(ipPortMapper));
  initIPServices();

  while ((op = getopt(argc, argv, "f:F:hr:i:p:l:nw:m:")) != EOF)
    switch (op) {

    case 'f':
      rFileName = optarg;
      break;

    case 'r':
      refreshRate = atoi(optarg);
      break;

    case 'i':
      device = optarg;
      break;

    case 'p':
      showProtoWarnings = 1;
      handleProtocols(optarg);
      break;

    case 'F':
      strcpy(flowSpecs, optarg); 
      break;

    case 'm':
      localAddresses = strdup(optarg);
      break;

    case 'l':
      logTimeout = atoi(optarg);
      if(logTimeout < 0) {
	printf("Log period out of range: set to default (30 mins)\n");
	logTimeout = 1800; /* seconds */
      }

      logd = fopen("ntop.log", "w+");
      if(logd == NULL) {
	fprintf(stderr, "Logging disabled: unable to log onto file ntop.log.\n");
	logTimeout = 0;
      } else {
	fprintf(logd, "# %s %s %s %s %s %s %s %s (Kbps)\n",
		"date", "totalPkts", "broadcastPkts", "multicastPkts",
		"ethernetBytes", "ipBytes", "nonIpBytes",
		"peakThroughput");
      }
      break;
      
    case 'n':
      numericFlag++;
      break;

    case 'w':
      webMode++;
      webPort = atoi(optarg);
      break;

    default:
      usage(0);
      exit(-1);
      /* NOTREACHED */
    }

  if(webMode) {
    /* 
       Make sure that the other flags are't set. They have
       no effect in web mode 
    */
    pflag = 0;
    percentMode = 0;

    if(refreshRate == ALARM_TIME /* default */)
      refreshRate = 120;
    else if(refreshRate < 30)
      refreshRate = 30;
  }
  
  if(numIpProtosToMonitor == 0)
    addDefaultProtocols();

  lastTime = time(NULL) + refreshRate;

#ifndef MULTITHREADED
  if(logTimeout != 0)
    nextLogTime = time(NULL) + logTimeout;
#endif

  /* Determine the device name if not specified */
  if (device == NULL && (device = pcap_lookupdev(ebuf)) == NULL)
    error(ebuf);

  getLocalHostAddress(&localHostAddress, device);

  if(rFileName == NULL) {
    if((getuid () && geteuid ()) || setuid (0)) {
      error("Sorry, you must be root in order to run this program.");
    }
    
  if(webMode) {
    (void)fprintf(stderr, "ntop v.%s %s [%s] listening on %s\n", 
		  version, THREAD_MODE, osName, device);
    (void)fprintf(stderr, "Copyright 1998-99 by %s\n", author);
    (void)fflush(stderr);
  } else
    init_curses();

    /* Fire up libpcap */
    pcapPtr = pcap_open_live(device, DEFAULT_SNAPLEN, !pflag, 1000, ebuf);
  } else {
    pcapPtr = pcap_open_offline(rFileName, ebuf);
  }

  if (pcapPtr == NULL)
    error(ebuf);

  if (pcap_lookupnet(device, &localnet, &netmask, ebuf) < 0)
    error(ebuf);

  localnet = htonl(localnet);
  netmask = htonl(netmask);

  if(localAddresses != NULL) {
    handleLocalAddresses(localAddresses);
    free(localAddresses);
    localAddresses = NULL;
  }

  /* Now that the socket is open, throw away potential setuid/setgid */
  (void)setgid(getgid());
  (void)setuid(getuid());

  /* get datalink type */
#ifdef AIX
  /* This is a bug of libpcap on AIX */
  switch(device[0]) {
  case 't': /* TokenRing */
    datalink = DLT_IEEE802;
    break;
  case 'l': /* Loopback */
    datalink = DLT_NULL;
    break;
  default:
    datalink = DLT_EN10MB; /* Ethernet */
  }
#else
  datalink = pcap_datalink(pcapPtr);
#endif

  /* Construct, compile and set filter */
  expression = copy_argv(argv + optind);
  if(expression != NULL) {
    if (pcap_compile(pcapPtr, &fcode, expression, 1, netmask) < 0 ||
	pcap_setfilter(pcapPtr, &fcode) < 0)
      error(pcap_geterr(pcapPtr));
  }

  /* Handle flows (if any) */
  if(flowSpecs[0] != '\0')
    handleFlowsSpecs(flowSpecs);

  /* Setup signal handlers */
  (void)setsignal(SIGTERM, cleanup);
  (void)setsignal(SIGINT, cleanup);

  if(webMode) 
    (void)setsignal(SIGPIPE, ignoreSignal); 

  /* Cooperate with nohup(1) */
  if ((oldhandler = setsignal(SIGHUP, cleanup)) != SIG_DFL)
    (void)setsignal(SIGHUP, oldhandler);

  init_counters();

#ifdef MULTITHREADED
  packetQueueLen = maxPacketQueueLen = packetQueueHead = packetQueueTail = 0;
  droppedPackets = 0;
#ifdef USE_SEMAPHORES
  createSem(&queueSem, 0);
#ifdef ASYNC_ADDRESS_RESOLUTION
  createSem(&queueAddressSem, 0);
#endif
#else
  createCondvar(&queueCondvar);
#ifdef ASYNC_ADDRESS_RESOLUTION
  createCondvar(&queueAddressCondvar);
#endif
#endif
  createMutex(&packetQueueMutex);
#ifdef HAVE_LSOF
  createMutex(&lsofMutex);
#endif
  createMutex(&hostsHashMutex);
  createThread(&dequeueThreadId, dequeuePacket);

  createThread(&thptUpdateThreadId, updateThptLoop);
  createThread(&scanIdleThreadId, scanIdleLoop);
  if(logTimeout != 0)
    createThread(&logFileLoopThreadId, logFileLoop);

  createThread(&lsofThreadId, periodicLsofLoop);

#ifdef ASYNC_ADDRESS_RESOLUTION
  addressQueueLen = maxAddressQueueLen = addressQueueHead = addressQueueTail = 0;
  droppedAddresses = 0;
  createMutex(&addressQueueMutex);
  createThread(&dequeueAddressThreadId, dequeueAddress);
#endif
#endif

  if(!webMode) {
#ifdef MULTITHREADED
    createThread(&cursesRefreshThreadId, handleCursesRefresh);
#else
    (void)setsignal(SIGALRM, _printHostsTraffic); 
#endif
  } else {
     (void)setsignal(SIGALRM, dontFreeze);
    /* 
       The handler below has been restored due to compatibility
       problems with om: 
       Courtesy of Martin Lucina <mato@kotelna.sk>
    */
    (void)setsignal(SIGCHLD, handleDiedChild);
    /* signal(SIGCHLD, SIG_IGN); */
  }
  

  if(!webMode)
    printHostsTraffic(-1, 0);
  else {
    int sockopt = 1;

#ifdef DEBUG
    numChildren = 0;
#endif    
    readConfigurationFile(); /* read ~/.ntop */

    sin.sin_family      = AF_INET;
    sin.sin_port        = htons(webPort);
    sin.sin_addr.s_addr = INADDR_ANY;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if(sock < 0)
      {
	error("unable to create a new socket");
	return(-1);
      }

    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&sockopt, sizeof(sockopt));

    if(bind(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
      sprintf(ebuf, "bind: port %d already in use.", webPort);
      error(ebuf);
      (void) close(sock);
      return(-1);
    } 
    
    if(listen(sock, 5) < 0) {
      sprintf(ebuf, "listen error.");
      error(ebuf);
      (void) close(sock);
      return(-1);
    } else {
      fprintf(stderr, "Waiting for HTTP connections on port %d...\n", webPort);
      (void)fflush(stderr);
    }
  }

#ifdef MULTITHREADED
  if(webMode) 
    createThread(&handleWebConnectionsThreadId, handleWebConnections);
#endif

  initialSniffTime = lastRefreshTime = time(NULL);

#ifdef HAVE_LSOF 
  read_lsofInfo(1);
#endif

  /* pcap_loop(pcapPtr, -1, pbuf_process, NULL); */
  for(;;) {    
#ifndef MULTITHREADED
    short justRefreshed;
#endif

#ifdef MULTITHREADED
    rc = pcap_dispatch(pcapPtr, -1, queuePacket, NULL);   
#else
    rc = pcap_dispatch(pcapPtr, -1, pbuf_process, NULL);   
#endif

    actTime = time(NULL);

#ifndef MULTITHREADED
    if(actTime > lastTime) {      
      if(!webMode) 
	printHostsTraffic(0, 0);
      else if(nextSessionTimeoutScan < actTime) {
	/* It's time to check for timeout sessions */
	scanTimedoutTCPSessions();
	purgeIdleHosts();
	nextSessionTimeoutScan = actTime+SESSION_SCAN_DELAY;
      }
      
      updateThpt(); /* Update Throughput */
      lastTime = actTime + refreshRate; /* refresh in 'refreshRate' seconds */
      justRefreshed=1;
    } else
      justRefreshed=0;
#endif

    if(!webMode) {
#ifndef MULTITHREADED
      if(checkKeyPressed()) {
	if(!justRefreshed) {
	  printHostsTraffic(0, 0);
	  lastTime = actTime + refreshRate; /* refresh in 'refreshRate' seconds */
	}
      }
#endif
    } else {
      /* webMode */
#ifndef MULTITHREADED
      handleWebConnections(NULL);
#endif
    }

#ifndef MULTITHREADED
    if((logTimeout != 0) && (actTime > nextLogTime)) {
      LogStatsToFile();
      nextLogTime = actTime + logTimeout;
    }
#endif
  }/* for(;;) */ 
    
  error("pcap_loop exited");
}

/* **************************************** */

void handleSingleWebConnection() {
  struct sockaddr_in from;
  int from_len = sizeof(from);
  newSock  = accept(sock, (struct sockaddr*)&from, &from_len);
	  
  if(newSock > 0) {
    int childpid;

#ifdef DONT_USE_FORK
    childpid = 0;
#else
    if((childpid=fork()) < 0) {
      printf("ntop: fork error [errno=%d]\n", errno);
      exit(0);
    } else
#endif
      {	    
	if(!childpid) {
	  /* webMode - child process */
#ifndef DONT_USE_FORK
	  /* There's no need to sniff anymore because 
	     ntop will quit after having served 
	     the HTTP client 
	  */
	  detachFromTerminal();
#endif
	  handleHTTPrequest(); 
#ifndef DONT_USE_FORK
	  pcap_close(pcapPtr);
#endif
	  close(newSock);
#ifndef DONT_USE_FORK
	  exit(0);
#endif
	} else {
	  /* webMode - parent process */
#ifdef DEBUG
	  int pid;
	  numChildren++;
	  printf("Forked a new child (pid=%d): %d\n", 
		 childpid, numChildren);
#endif
	  close(newSock);
	  /* wait(&pid); */
	}
      }
  }
}


/* **************************************** */

void* handleWebConnections(void* notUsed) {
#ifndef MULTITHREADED  
  struct timeval wait_time;
#endif
  fd_set mask;

  FD_ZERO(&mask);
  FD_SET(sock, &mask);    

#ifndef MULTITHREADED  
  /* select returns immediately */
  wait_time.tv_sec = 0, wait_time.tv_usec = 0; 
  if(select(sock+1, &mask, 0, 0, &wait_time) == 1)
    handleSingleWebConnection();
#else
  while(1) {
    if(select(sock+1, &mask, 0, 0, NULL /* Infinite */) == 1)
      handleSingleWebConnection();
  }

#endif

  return(NULL); /* NOTREACHED */
}


/* **************************************** */

RETSIGTYPE handleDiedChild(int signal) {
  int status;
  pid_t pidId;

  pidId = waitpid(-1, &status, WNOHANG);

#ifdef DEBUG
  if(status == 0) {
    numChildren--;
    printf("A child has terminated [pid=%d status=%d children=%d]\n",
	   pidId, status, numChildren);
  }
#endif

  (void)setsignal(SIGCHLD, handleDiedChild);
}

/* **************************************** */

RETSIGTYPE dontFreeze(int signo)
{
  printf("Child timed out. Bye bye...\n");
  exit(0);
}

/* **************************************** */

/* Report statistics and write out the raw packet file */
RETSIGTYPE cleanup(int signo)
{
  struct pcap_stat stat;

#ifdef MULTITHREADED
  /* Courtesy of Felipe Tonioli <tonioli@mtec.com.br> */
  if(signo != -1) /* the user pressed the 'q' key */
    {
      killThread(&dequeueThreadId);
      killThread(&thptUpdateThreadId);
      killThread(&scanIdleThreadId);
      killThread(&lsofThreadId);
#ifdef ASYNC_ADDRESS_RESOLUTION
      killThread(&dequeueAddressThreadId);
#endif

      if(!webMode) 
	killThread(&cursesRefreshThreadId);
      else
	killThread(&handleWebConnectionsThreadId);

      if(logTimeout != 0)
	killThread(&logFileLoopThreadId);

      deleteMutex(&packetQueueMutex);
      deleteMutex(&hostsHashMutex);
#ifdef HAVE_LSOF
      deleteMutex(&lsofMutex);
#endif
#ifdef USE_SEMAPHORES
      deleteSem(&queueSem);
#ifdef ASYNC_ADDRESS_RESOLUTION
      deleteSem(&queueAddressSem);    
#endif
#else
      deleteCondvar(&queueCondvar);
#ifdef ASYNC_ADDRESS_RESOLUTION
      deleteCondvar(&queueAddressCondvar);    
#endif
#endif
    }
#endif

  if(!webMode)
    cleanup_curses();

  (void)fflush(stdout);

  if (pcap_stats(pcapPtr, &stat) < 0)
    (void)fprintf(stderr, "\npcap_stats: %s\n", pcap_geterr(pcapPtr));
  else {
    (void)fprintf(stderr, "%s packets received by filter\n",
		  formatPkts((TrafficCounter)stat.ps_recv));
    (void)fprintf(stderr, "%s packets dropped by kernel\n",
		  formatPkts((TrafficCounter)(stat.ps_drop)));
#ifdef MULTITHREADED
    (void)fprintf(stderr, "%s packets dropped by ntop\n",
		  formatPkts(droppedPackets));
#endif
    if (logTimeout) {
      fclose(logd);
    }
  }
  
  endservent();

  exit(0);
}

/* **************************************** */

void usage(int cursesMode)
{
  char buf[80];
  int row=0;

  printingUsage = 1;

  if(cursesMode) {
#ifndef MULTITHREADED
    alarm(0); /* Avoid to be interrupted */
#endif
    clrscr(); /* clear the screen */  
  }

  sprintf(buf, "%s v.%s for %s", program_name, version, osName);
  if(cursesMode) { mvprintw (row++, 0, buf); } else fprintf(stderr, "%s\n", buf);
  sprintf(buf, "Written by %s.", author);
  if(cursesMode) { mvprintw (row, 0, buf); row+=2; } else fprintf(stderr, "%s\n\n", buf);
  sprintf(buf, "Usage: %s", program_name);
  if(!cursesMode) fprintf(stderr, "%s\n", buf);

  if(!cursesMode) {
    fprintf(stderr,"    %s\n",   "[-r <refresh time (interactive = 3 sec./web = 120 sec.)>]");
    fprintf(stderr,"    %s\n",   "[-f <traffic dump file (see tcpdump)>]");
    fprintf(stderr,"    %s\n",   "[-n (numeric IP addresses (interactive mode only))]");
    fprintf(stderr,"    %s\n",   "[-p <IP protocols to monitor> (see man page)]");
    fprintf(stderr,"    %s\n",   "[-i <interface>]");
    fprintf(stderr,"    %s\n",   "[-w <port>]");  
    fprintf(stderr,"    %s\n",   "[-m <local addresses (see man page)>]");  
    fprintf(stderr,"    %s\n",   "[-l <log period (seconds)>]");  
    fprintf(stderr,"    %s\n",   "[-F <flow specs (see man page)>]");  
    fprintf(stderr,"    %s\n\n", "[ <filter expression (like tcpdump)>]");
  }

  sprintf(buf, "While %s is running interactively, the following keys are active:", program_name);
  if(cursesMode) { mvprintw (row++, 0, buf); } else fprintf(stderr, "%s\n", buf);
  sprintf(buf, "'q' - quit ntop");
  if(cursesMode) { mvprintw (row++, 7, buf); } else fprintf(stderr, "    %s\n", buf);
  sprintf(buf, "'r' - reset statistics");
  if(cursesMode) { mvprintw (row++, 7, buf); } else fprintf(stderr, "    %s\n", buf);
  sprintf(buf, "'n' - toggle address format (num <-> sym <-> MAC <-> Nw Board Manifact.)");
  if(cursesMode) { mvprintw (row++, 7, buf); } else fprintf(stderr, "    %s\n", buf);
  sprintf(buf, "'p' - toggle traffic values (bytes <-> %% <-> thpt)");
  if(cursesMode) { mvprintw (row++, 7, buf); } else fprintf(stderr, "    %s\n", buf);
  sprintf(buf, "'l' - toggle hosts display (local subnet <-> all)");
  if(cursesMode) { mvprintw (row++, 7, buf); } else fprintf(stderr, "    %s\n", buf);
  sprintf(buf, "'d' - toggle idle (idle <-> send/receive)");
  if(cursesMode) { mvprintw (row++, 7, buf); } else fprintf(stderr, "    %s\n", buf);
  sprintf(buf, "'t' - toggle sort (sent <-> received)");
  if(cursesMode) { mvprintw (row++, 7, buf); } else fprintf(stderr, "    %s\n", buf);
  sprintf(buf, "'y' - toggle columns sort");
  if(cursesMode) { mvprintw (row++, 7, buf); } else fprintf(stderr, "    %s\n", buf);
  sprintf(buf, "'h' - show this help");
  if(cursesMode) { mvprintw (row++, 7, buf); } else fprintf(stderr, "    %s\n", buf);
  sprintf(buf, "' ' - toggle protocol");
  if(cursesMode) { mvprintw (row++, 7, buf); } else fprintf(stderr, "    %s\n", buf);

  if(cursesMode) { 
    mvprintw(++row, 0, "Strike a key to continue..."); 
    refresh();

    while(getchar() == EOF)
      ;

#ifndef MULTITHREADED
    alarm(refreshRate);
#endif
  }

  printingUsage = 0;
}

/* **************************************** */

void detachFromTerminal() {
  int fd;
  int i;

  return ;

#ifndef MULTITHREADED
  alarm(120); /* Don't freeze */
#endif

  setsid();  /* detach from the terminal */

  close(0); close(1); close(2);
  /* try and use up these file descriptors, so silly
     library routines writing to stdout etc won't cause havoc */

#ifdef HAVE_SYS_FCNTL_H
  for (i=0;i<3;i++) {
    fd = open("/dev/null",O_RDWR,0);
    if (fd < 0) fd = open("/dev/null",O_WRONLY,0);
    if (fd < 0) {
      return;
    }
    if (fd != i) {
      return;
    }
  }
#endif
}

/* **************************************** */

void handleProtocols(char *protos) {
  char *proto = strtok(protos, ",");

  while(proto != NULL) {
    char* protoName = strchr(proto, '=');

    if(protoName == NULL)
      printf("Unknown protocol '%s'. It has been ignored.\n", proto);
    else {
      char tmpStr[255];
      int len;

      protoName[0] = '\0';
      memset(tmpStr, 0, sizeof(tmpStr));
      strcpy(tmpStr, &protoName[1]);
      len = strlen(tmpStr);

      if(tmpStr[len-1] != '|') {
	/* Make sure that the string ends with '|' */
	tmpStr[len] = '|';
	tmpStr[len+1] = '\0';
      }

      handleProtocolList(proto, tmpStr);
      proto = strtok(NULL, ",");
    }
  }
}

/* **************************************** */

void addDefaultProtocols() {
  handleProtocolList("FTP", "ftp|ftp-data|");
  handleProtocolList("HTTP", "http|www|https|");
  handleProtocolList("DNS", "name|");
  handleProtocolList("Telnet", "telnet|");
  handleProtocolList("NBios-IP", "netbios-ns|netbios-dgm|netbios-ssn|");
  handleProtocolList("POP", "pop-2|pop-3|pop3|kpop|");
  handleProtocolList("SNMP", "snmp|snmp-trap|");
  handleProtocolList("NFS", "mount|pcnfs|bwnfs|nfsd|nfsd-status|");
  handleProtocolList("X11", "6000-6010|");
}

/* **************************************** */

short handleProtocol(char* protoName, char *protocol) {
  int i, idx, lowProtoPort, highProtoPort;
  
  if(protocol[0] == '\0')
    return(-1);
  else if(isdigit(protocol[0])) {
    /* numeric protocol port handling */
    lowProtoPort = highProtoPort = 0;
    sscanf(protocol, "%d-%d", &lowProtoPort, &highProtoPort);
    if(highProtoPort < lowProtoPort)
      highProtoPort = lowProtoPort;

    if(lowProtoPort < 0) lowProtoPort = 0;
    if(highProtoPort >= TOP_IP_PORT) highProtoPort = TOP_IP_PORT-1;

    for(idx=lowProtoPort; idx<= highProtoPort; idx++) {
      if(ipPortMapper[idx] == -1) {
	numIpPortsToHandle++;
	
#ifdef DEBUG
	printf("[%d] '%s' [port=%d]\n", numIpProtosToMonitor, protoName, idx);
#endif
	ipPortMapper[idx] = numIpProtosToMonitor;
      } else if(showProtoWarnings)
	printf("WARNING: IP port %d (%s) has been discarded (multiple instances).\n", 
	       idx, protoName);
    }

    return(1);
  } 

  for(i=1; i<SERVICE_HASH_SIZE; i++) {
    idx = -1;

    if((udpSvc[i] != NULL) && (strcmp(udpSvc[i]->name, protocol) == 0))
      idx = udpSvc[i]->port;
    else if((tcpSvc[i] != NULL) && (strcmp(tcpSvc[i]->name, protocol) == 0))
      idx = tcpSvc[i]->port;

    if(idx != -1) {
      if(ipPortMapper[idx] == -1) {
	numIpPortsToHandle++;

#ifdef DEBUG
	printf("[%d] '%s' [%s:%d]\n", numIpProtosToMonitor, protoName, protocol, idx);
#endif
	ipPortMapper[idx] = numIpProtosToMonitor;
      } else if(showProtoWarnings)
	printf("WARNING: protocol '%s' has been discarded (multiple instances).\n", protocol);
      
      return(1);
    }
  }

  if(showProtoWarnings)
    printf("WARNING: unknown protocol '%s'. It has been ignored.\n", protocol);

  return(-1);
}

/* **************************************** */

void handleProtocolList(char* protoName, char *protocolList) {
  char tmpStr[255];
  char* lastEntry, *protoEntry;
  int increment=0, rc;

#ifdef DEBUG
  printf("%s - %s\n", protoName, protocolList);
#endif

  if(numIpProtosToMonitor == MAX_NUM_HANDLED_IP_PROTOCOLS) {
    /* MAX_NUM_HANDLED_IP_PROTOCOLS is defined in ntop.h: increase it! */
    printf("WARNING: Unable to handle '%s'. You've reached the max number\n"
	   "of handled IP protocols.\n", protoName);
    return;
  }

  /* The trick below is used to avoid to modify static
     memory like in the case where this function is
     called by addDefaultProtocols()
  */
  lastEntry = strcpy(tmpStr, protocolList);

  while((protoEntry  = strchr(lastEntry, '|')) != NULL) {
    protoEntry[0] = '\0';
    rc = handleProtocol(protoName, lastEntry);

    if(rc != -1)
      increment=1;

    lastEntry = &protoEntry[1];
  }

  if(increment == 1) {
    protoIPTrafficInfos[numIpProtosToMonitor] = strdup(protoName);    
    numIpProtosToMonitor++;
#ifdef DEBUG
    printf("%d) %s - %s\n", numIpProtosToMonitor, protoName, protocolList); 
#endif
  }
}

/* **************************************** */

int mapGlobalToLocalIdx(int port) {
  if((port < 0) || (port >= TOP_IP_PORT))
    return -1;
  else {
#ifdef DEBUG
    printf("IP port %d -> %d\n", port, ipPortMapper[port]);
#endif
    return(ipPortMapper[port]);
  }
}

/* **************************************** */

#ifdef MULTITHREADED
void* updateThptLoop(void* notUsed) {

  for(;;) {
    /* printf("Sleeping for %d seconds\n", refreshRate); */
    sleep(refreshRate);
    /* Don't update Thpt if the traffic is high */

    /* printf("Trying to update throughput\n"); */

    if(packetQueueLen < (PACKET_QUEUE_LENGTH/3)) {
      actTime = time(NULL);
#ifdef MULTITHREADED
      accessMutex(&hostsHashMutex);
#endif
      /* printf("Updating throughput\n"); */
      updateThpt(); /* Update Throughput */
#ifdef MULTITHREADED
      releaseMutex(&hostsHashMutex);
#endif
    }
  }
}


void* scanIdleLoop(void* notUsed) {
  for(;;) {
    sleep(SESSION_SCAN_DELAY);   
    actTime = time(NULL);

    /* Don't purge hosts if the traffic is high */
    if(packetQueueLen < (PACKET_QUEUE_LENGTH/3)) {
#ifdef MULTITHREADED
      accessMutex(&hostsHashMutex);
#endif      
      scanTimedoutTCPSessions();
      purgeIdleHosts();
#ifdef MULTITHREADED
      releaseMutex(&hostsHashMutex);
#endif
    }
  }
}

void* logFileLoop(void* notUsed) {
  for(;;) {
    sleep(logTimeout);
    actTime = time(NULL);

    /* Don't purge hosts if the traffic is high */
    if(packetQueueLen < (PACKET_QUEUE_LENGTH/3)) {
#ifdef MULTITHREADED
      accessMutex(&hostsHashMutex);
#endif      
      LogStatsToFile();
#ifdef MULTITHREADED
      releaseMutex(&hostsHashMutex);
#endif
    }
  }
}

#ifdef MULTITHREADED
#ifdef HAVE_LSOF
void* periodicLsofLoop(void* notUsed) {
  for(;;) {
    /* refresh process list each minute */
    sleep(60);
    read_lsofInfo(0);
  }
}
#endif
#endif

#endif

