/*
 *  Copyright (C) 1998-99 Luca Deri <deri@unipi.it>
 *                        Stefano Suin <stefano@unipi.it>
 *                        (Stefano contributed to version 1.0)
 *                      
 *  			  Centro SERRA, University of Pisa
 *  			  http://www-serra.unipi.it/
 *  					
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Copyright (c) 1994, 1996
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */


#include "ntop.h"

/* #define DEBUG */

/* #define DEBUG_THREADS */

/* Extern */
extern int newSock, datalink;
extern struct in_addr localHostAddress;
extern u_short numIpProtosToMonitor;
extern void sendString(char *x);
extern void usage(int);
extern void createVendorTable();
extern char* getVendorInfo(u_char* ethAddress);
extern unsigned int localnet, netmask, webMode;
extern RETSIGTYPE cleanup(int signo);
extern int numericFlag, refreshRate, localAddrFlag, percentMode, idleFlag, logFlag;
extern char *device;
extern FILE *logd;
extern short screenNumber;
extern u_short columnSort;
extern void printHeader(int);
extern void printHelp();
extern void error(const char *fmt, ...);
extern void printSession(IPSession *theSession, u_short sessionType, 
			 u_short sessionCounter);
extern short isLocalAddress(struct in_addr *addr);
extern short isPseudoLocalAddress(struct in_addr *addr);
extern short isBroadcastAddress(struct in_addr *addr);
extern int32_t gmt2local(time_t t);
extern int mapGlobalToLocalIdx(int port);
extern short isMulticastAddress(struct in_addr *addr);
extern time_t actTime, initialSniffTime;
extern TrafficCounter lastTotalPkts, lastBroadcastPkts, lastMulticastPkts,
  lastEthernetBytes, lastIpBytes, lastNonIpBytes;
extern FlowFilterList *flowsList;
#ifdef MULTITHREADED
extern pthread_mutex_t     packetQueueMutex, hostsHashMutex;
extern pthread_t           dequeueThreadId;
extern PacketInformation packetQueue[PACKET_QUEUE_LENGTH];
extern unsigned int packetQueueLen, maxPacketQueueLen, packetQueueHead, packetQueueTail;
extern TrafficCounter droppedPackets;
#ifdef USE_SEMAPHORES
extern sem_t queueSem;
#ifdef ASYNC_ADDRESS_RESOLUTION
extern sem_t queueAddressSem;
#endif
#else
extern ConditionalVariable queueCondvar;
#ifdef ASYNC_ADDRESS_RESOLUTION
extern ConditionalVariable queueAddressCondvar;
#endif
#endif

#ifdef ASYNC_ADDRESS_RESOLUTION
extern unsigned int addressQueueLen, maxAddressQueueLen, addressQueueHead, addressQueueTail;
extern TrafficCounter droppedAddresses;
extern pthread_mutex_t addressQueueMutex;
extern struct hnamemem *addressQueue[ADDRESS_QUEUE_LENGTH];
#endif
#endif

#ifdef HAVE_LSOF /* util.c */
extern ProcessInfo *processes[MAX_NUM_PROCESSES];
extern u_short numProcesses;
extern ProcessInfo *localPorts[TOP_IP_PORT][MAX_NUM_SHARED_PORTS];
#endif


/* Global */
u_char dummyEthAddress[ETHERNET_ADDRESS_LEN];
short sortSendMode=0;
int lastNumLines, lastNumCols;
time_t nextSessionTimeoutScan;
struct timeval lastPktTime;
PacketStats rcvdPktStats;
SimpleProtoTrafficInfo tcpGlobalTrafficStats, udpGlobalTrafficStats;

/* Static */
static time_t lastThptUpdate, lastMinThptUpdate, lastFiveMinsThptUpdate;
static char domainName[MAXHOSTNAMELEN];
static char hex[] = "0123456789ABCDEF";
static TrafficCounter throughput;
static short alternateColor = 0, maxNameLen;
static HostTraffic broadcastEntry;
static TrafficCounter hlen;

/* Global */
jmp_buf getname_env;
float peakThroughput, actualThpt, lastMinThpt, lastFiveMinsThpt;
struct hnamemem* hnametable[HASHNAMESIZE];
struct enamemem enametable[HASHNAMESIZE];
struct hnamemem eprototable[HASHNAMESIZE];
IpFragment *fragmentTable[HASHNAMESIZE];
HostTraffic* hash_hostTraffic[HASHNAMESIZE];
u_short	extracted_ethertype;
TrafficCounter ethernetPkts, ethernetBytes, ipBytes; 
TrafficCounter broadcastPkts, multicastPkts;
TrafficCounter lastMinEthernetBytes, lastFiveMinsEthernetBytes;
IPSession *tcpSession[HASHNAMESIZE]; /* TCP sessions */
IPSession *udpSession[HASHNAMESIZE]; /* UDP sessions */
u_short numTcpSessions=0, numUdpSessions=0;
IpGlobalSession *Gsessions[HASHNAMESIZE]; /* Sessions Global Information  */
char *separator;
ServiceEntry *udpSvc[SERVICE_HASH_SIZE], *tcpSvc[SERVICE_HASH_SIZE];
TrafficEntry ipTrafficMatrix[256][256]; /* Subnet traffic Matrix */
HostTraffic* ipTrafficMatrixHosts[256]; /* Subnet traffic Matrix Hosts */
int32_t thisZone; /* seconds offset from gmt to local time */
SimpleProtoTrafficInfo *ipProtoStats;
TrafficCounter tcpBytes, udpBytes, icmpBytes, dlcBytes, ipxBytes, decnetBytes, netbiosBytes, arpRarpBytes, atalkBytes, otherBytes, ospfBytes, egpBytes, igmpBytes, otherIpBytes;

/* Forward */
char* ipaddr2str(struct in_addr hostIpAddress);
char* savestr(const char *str);
RETSIGTYPE printHostsTraffic(int signumber_ignored, int);
char* intoa(struct in_addr addr);
void updateThpt();
void purgeIdleHosts();
char* etheraddr_string(const u_char *ep);
void updatePacketCount(HostTraffic *srcHost, HostTraffic *dstHost, TrafficCounter length);
void updateTrafficMatrix(HostTraffic *srcHost, HostTraffic *dstHost, TrafficCounter length);
static void freeHostInfo(HostTraffic *host);
void purgeOldFragmentEntries();
#if defined(MULTITHREADED) && defined(ASYNC_ADDRESS_RESOLUTION)
void queueAddress(struct hnamemem* elem);
#endif

/* ************************************ */

void clrscr()
{
  attrset (A_NORMAL);
  erase();

  /* border(0, 0, 0, 0, 0, 0, 0, 0); */
  /* refresh(); */
}

/* ******************************* */

char* getRowColor() {
#define USE_COLOR

#ifdef USE_COLOR
  if(alternateColor == 0) {
    alternateColor = 1;
    return("BGCOLOR=#FFFFCC");
  } else {
    alternateColor = 0;
    return("");
  }
#else
  return("");
#endif
}

/* ******************************* */

char* getActualRowColor() {
#define USE_COLOR

#ifdef USE_COLOR
  if(alternateColor == 1)
    return("BGCOLOR=#FFFFCC");
  else 
    return("");
#else
  return("");
#endif
}

/* ******************************* */

void init_curses ()
{
  if(webMode)
    return;

  initscr();
  cbreak();
  noecho();
  nodelay(stdscr, TRUE);
  nonl();
  intrflush(stdscr, FALSE);
  keypad(stdscr, TRUE);
 
  attrset(A_NORMAL);

  lastNumLines = LINES;
  lastNumCols = COLS;

  printHeader(0);
}

/* ******************************* */

void cleanup_curses ()
{
  if(webMode) return;
  endwin ();
}

/* ******************************* */

void resetStats() {
  int idx;

  ethernetPkts = 0;
  ethernetBytes = 0;
  ipBytes = 0;
  broadcastPkts = 0;
  multicastPkts = 0;
  peakThroughput = 0;
  throughput = 0;
  lastThptUpdate = lastMinThptUpdate = lastFiveMinsThptUpdate = time(NULL);
  lastMinEthernetBytes = lastFiveMinsEthernetBytes = 0;

  for(idx=0; idx<HASHNAMESIZE; idx++) {
    memset(&hash_hostTraffic[idx], 0, sizeof(hash_hostTraffic[idx]));
  }

  memset(&tcpGlobalTrafficStats, 0, sizeof(tcpGlobalTrafficStats));
  memset(&udpGlobalTrafficStats, 0, sizeof(udpGlobalTrafficStats));
}

/* ******************************* */

int getPortByName(ServiceEntry **theSvc, char* portName) {
  int idx;

  for(idx=0; idx<SERVICE_HASH_SIZE; idx++) {

#ifdef DEBUG
    if(theSvc[idx] != NULL) 
      printf("%d/%s [%s]\n", 
	     theSvc[idx]->port, 
	     theSvc[idx]->name, portName);
#endif

    if((theSvc[idx] != NULL) 
       && (strcmp(theSvc[idx]->name, portName) == 0))
      return(theSvc[idx]->port);
  }

  return(-1);
}

/* ******************************* */

char* getPortByNumber(ServiceEntry **theSvc, int port) {
  int idx = port % SERVICE_HASH_SIZE;
  ServiceEntry *scan;

  for(;;) {
    scan = theSvc[idx];

    if((scan != NULL) && (scan->port == port))
      return(scan->name);
    else if(scan == NULL)
      return(NULL);
    else
      idx = (idx+1) % SERVICE_HASH_SIZE;
  }
}

/* ******************************* */

char* getPortByNum(int port, int type) {
  char* rsp;

  if(type == IPPROTO_TCP) {
    rsp = getPortByNumber(tcpSvc, port);
  } else {
    rsp = getPortByNumber(udpSvc, port);
  }

  return(rsp);
}

/* ******************************* */

char* getAllPortByNum(int port) {
  char* rsp;

  rsp = getPortByNumber(tcpSvc, port); /* Try TCP first... */
  if(rsp == NULL)
    rsp = getPortByNumber(udpSvc, port);  /* ...then UDP */
  
  return(rsp);
}

/* ******************************* */

int getAllPortByName(char* portName) {
  int rsp;

  rsp = getPortByName(tcpSvc, portName); /* Try TCP first... */
  if(rsp == -1)
    rsp = getPortByName(udpSvc, portName);  /* ...then UDP */
  
  return(rsp);
}


/* ******************************* */

void addPortHashEntry(ServiceEntry **theSvc, int port, char* name) {
  int idx = port % SERVICE_HASH_SIZE;
  ServiceEntry *scan;

  for(;;) {
    scan = theSvc[idx];

    if(scan == NULL) {
      theSvc[idx] = (ServiceEntry*)malloc(sizeof(ServiceEntry));
      theSvc[idx]->port = (u_short)port;
      theSvc[idx]->name = strdup(name);
      break;
    } else
      idx = (idx+1) % SERVICE_HASH_SIZE;
  }
}

/* ******************************* */

RETSIGTYPE windowSizeChanged(int signumber_ignored) {
  lastNumLines = 0; /* force redraw */
  printHostsTraffic(signumber_ignored, 0);
}

/* ******************************* */

void initIPServices() {
  struct servent *svnt;

  memset(udpSvc, 0, sizeof(udpSvc));
  memset(tcpSvc, 0, sizeof(tcpSvc));

  setservent(1); /* used by getservbyXXX */

  while(((svnt = getservent()) != NULL) 
	&& (svnt->s_port < TOP_IP_PORT))
    /* Courtesy of Bernhard Heibler <bernhard_h@sz-testsys.de> */
    {
    if(strcmp(svnt->s_proto, "tcp") == 0)
      addPortHashEntry(tcpSvc, (int)ntohs((unsigned short int)svnt->s_port), svnt->s_name);
    else
      addPortHashEntry(udpSvc, (int)ntohs((unsigned short int)svnt->s_port), svnt->s_name);
  }

  setservent(0); /* used by getservbyXXX */

  /* Add some basic services, just in case they
     are not included in /etc/services */
  addPortHashEntry(tcpSvc, 21,  "ftp");
  addPortHashEntry(tcpSvc, 20,  "ftp-data");
  addPortHashEntry(tcpSvc, 80,  "http");
  addPortHashEntry(tcpSvc, 443, "https");
  addPortHashEntry(tcpSvc, 42,  "name");
  addPortHashEntry(tcpSvc, 23,  "telnet");
  addPortHashEntry(udpSvc, 137, "netbios-ns");
  addPortHashEntry(tcpSvc, 137, "netbios-ns");
  addPortHashEntry(udpSvc, 138, "netbios-dgm");
  addPortHashEntry(tcpSvc, 138, "netbios-dgm");
  addPortHashEntry(udpSvc, 139, "netbios-ssn");
  addPortHashEntry(tcpSvc, 139, "netbios-ssn");
  addPortHashEntry(tcpSvc, 109, "pop-2");
  addPortHashEntry(tcpSvc, 110, "pop-3");
  addPortHashEntry(tcpSvc, 1109,"kpop");
  addPortHashEntry(udpSvc, 161, "snmp");
  addPortHashEntry(udpSvc, 162, "snmp-trap");
  addPortHashEntry(udpSvc, 635, "mount");
  addPortHashEntry(udpSvc, 640, "pcnfs");
  addPortHashEntry(udpSvc, 650, "bwnfs");
  addPortHashEntry(udpSvc, 2049,"nfsd");
  addPortHashEntry(udpSvc, 1110,"nfsd-status");
}

/* ******************************* */

void init_counters() {
  char *p;

  /* (void)setsignal(SIGWINCH, windowSizeChanged);  */

  /* 
   * The name of the local domain is now calculated properly
   * Kimmo Suominen <kim@tac.nyc.ny.us> 
   */
  if((getdomainname(domainName, MAXHOSTNAMELEN) != 0) 
     || (domainName[0] == '\0'))
    {
      if ((gethostname(domainName, MAXHOSTNAMELEN) == 0)
	  && ((p = memchr(domainName, '.', MAXHOSTNAMELEN)) != NULL)) {
	domainName[MAXHOSTNAMELEN - 1] = '\0';
	/* 
	 * Replaced memmove with memcpy 
	 * Igor Schein <igor@txc.com>
	*/
	memcpy(domainName, ++p, (MAXHOSTNAMELEN+domainName-p));
      } else
	domainName[0] = '\0';
    }

  resetStats();
  createVendorTable();
  columnSort = 0;

  if(webMode) {
    separator = "&nbsp;";
    maxNameLen = MAXHOSTNAMELEN;
  } else {
    separator = " ";
    maxNameLen = MAX_HOST_NAME_LEN;
  }

  memset(dummyEthAddress, 0, ETHERNET_ADDRESS_LEN);
  memset(hnametable, 0, sizeof(hnametable));
  memset(enametable, 0, sizeof(enametable));
  memset(eprototable, 0, sizeof(eprototable));
  memset(hash_hostTraffic, 0, sizeof(hash_hostTraffic));
  memset(tcpSession, 0, sizeof(tcpSession));
  memset(udpSession, 0, sizeof(udpSession));
  memset(Gsessions, 0, sizeof(Gsessions));

  nextSessionTimeoutScan = time(NULL)+SESSION_SCAN_DELAY;
  thisZone = gmt2local(0);

  lastTotalPkts = lastBroadcastPkts = lastMulticastPkts = 0;
  lastEthernetBytes = lastIpBytes = lastNonIpBytes = 0;
  memset(&rcvdPktStats, 0, sizeof(rcvdPktStats));
  ipProtoStats = (SimpleProtoTrafficInfo*)calloc((size_t)numIpProtosToMonitor,
						 sizeof(SimpleProtoTrafficInfo));
  tcpBytes = udpBytes = icmpBytes = dlcBytes = ipxBytes = netbiosBytes = 
    decnetBytes = arpRarpBytes = atalkBytes = otherBytes = otherIpBytes = 0;

  broadcastEntry.broadcastHost = 
    broadcastEntry.subnetPseudoLocalHost = 
    broadcastEntry.subnetLocalHost = 1;

  memset(fragmentTable, 0, sizeof(fragmentTable));

  hlen = (datalink == DLT_NULL) ? NULL_HDRLEN : sizeof(struct ether_header);
}

/* ******************************* */

HostTraffic* getHostInfo(struct in_addr *hostIpAddress, 
			 u_char *ether_addr) 
{
  u_int idx, i, run=0;
  HostTraffic *el;
  short numCmp;

  if((hostIpAddress != NULL) && (!isLocalAddress(hostIpAddress))) {
    idx = 0;
    memcpy(&idx, &hostIpAddress->s_addr, 4);
    numCmp = 1;
    /* printf("-> %s\n", intoa(*hostIpAddress)); */
  } else if((hostIpAddress != NULL) && isBroadcastAddress(hostIpAddress)) {
    return(&broadcastEntry);
  } else {
    memcpy(&idx, &ether_addr[ETHERNET_ADDRESS_LEN-sizeof(u_int)], sizeof(u_int));
    numCmp = 0;
  }

  idx = (u_int)((idx*3) % HASHNAMESIZE);

 GET_HOST_INFO_AGAIN:
  for(i=0; i<HASHNAMESIZE; i++) {
    el = hash_hostTraffic[idx]; /* (**) */

    if(el != NULL) { 
      if(numCmp == 0) {
	if (memcmp(el->ethAddress, ether_addr, ETHERNET_ADDRESS_LEN) == 0) {
	  if((el->hostNumIpAddress[0] == '\0') && (hostIpAddress != NULL)) {
	    /* This entry didn't have IP fields set: let's set them now */
	    el->hostIpAddress.s_addr = hostIpAddress->s_addr;
	    strcpy(el->hostNumIpAddress, intoa(*hostIpAddress));
	    el->broadcastHost = isBroadcastAddress(&el->hostIpAddress);
	  }	
	  break;
	}
      } else {
	if (el->hostIpAddress.s_addr == hostIpAddress->s_addr)
	  break;
      }
    } else if(el == NULL) { 
      /* New table entry */
      el = (HostTraffic*)calloc(1, sizeof(HostTraffic));
      el->protoIPTrafficInfos = (ProtoTrafficInfo*)calloc((size_t)numIpProtosToMonitor,
							  sizeof(ProtoTrafficInfo));
      hash_hostTraffic[idx] = el;

      if((hostIpAddress == NULL)
	 || ((hostIpAddress != NULL) && isLocalAddress(hostIpAddress))) {
	/* This is a local address and then the 
	   ethernet address does make sense */
	memcpy(el->ethAddress, ether_addr, ETHERNET_ADDRESS_LEN);
	strcpy(el->ethAddressString, etheraddr_string(ether_addr));
	el->subnetLocalHost = LOCAL_HOST;
	el->subnetPseudoLocalHost = LOCAL_HOST;
      } else if(hostIpAddress != NULL) {
	/* This is packet that's being routed */
	memcpy(el->ethAddress, &hostIpAddress->s_addr, 4); /* Dummy/unique eth address */
	el->subnetLocalHost = REMOTE_HOST;
	if(isPseudoLocalAddress(hostIpAddress))
	  el->subnetPseudoLocalHost = LOCAL_HOST;
	else
	  el->subnetPseudoLocalHost = REMOTE_HOST;
      } else
	el->subnetLocalHost = REMOTE_HOST;

      if(strncmp(el->ethAddressString, "FF:", 3) == 0) {
	/* The trick below allows me not to duplicate the 
	   "<broadcast>" string in the code
	*/
	el->hostIpAddress.s_addr = INADDR_BROADCAST;
	el->broadcastHost = 1;
	strcpy(el->hostNumIpAddress, intoa(el->hostIpAddress));
	el->hostSymIpAddress = el->hostNumIpAddress;
      } else if(hostIpAddress != NULL) {
	el->hostIpAddress.s_addr = hostIpAddress->s_addr;
	strcpy(el->hostNumIpAddress, intoa(*hostIpAddress));
	el->broadcastHost = isBroadcastAddress(&el->hostIpAddress);
	/* Trick to fill up the address cache */
	if(numericFlag == 0)
	  el->hostSymIpAddress = ipaddr2str(el->hostIpAddress); 
	else
	  el->hostSymIpAddress = el->hostNumIpAddress;
      } else
	el->hostSymIpAddress = el->hostNumIpAddress;
      break;
    } 

    idx = (idx+1) % HASHNAMESIZE;
  }

  if(i == HASHNAMESIZE) {
    /* The hashtable is full */
    if(run == 0) {
      purgeIdleHosts();
    } else {
      /* No space yet: let's delete the oldest table entry */ 
      int candidate = 0;
      time_t lastSeenCandidate=0;
      HostTraffic* hostToFree;
      
      for(i=0; i<HASHNAMESIZE; i++)
	if(hash_hostTraffic[i] != NULL) {
	  if((candidate == 0) 
	     || (hash_hostTraffic[idx]->lastSeen < lastSeenCandidate)) {
	    candidate = i;
	    lastSeenCandidate = hash_hostTraffic[idx]->lastSeen;
	  }
	}

      hostToFree = hash_hostTraffic[candidate];
      hash_hostTraffic[candidate] = NULL;
      freeHostInfo(hostToFree);
      idx = candidate; /* this is a hint for (**) */
    }

    run++;      
  }

  el->lastSeen = actTime;

  return(el);
}

/* ******************************* */

char* ipaddr2str(struct in_addr hostIpAddress) {
  unsigned int addr = hostIpAddress.s_addr;
  int idx = addr % HASHNAMESIZE;
  struct hnamemem *p;
  struct hostent *hp = NULL;

  /* printf("Searching for %s.\n", intoa(hostIpAddress));  */

  if((addr == INADDR_BROADCAST) || (addr == 0x0))
    return(intoa(hostIpAddress));

  for(;;) {
    p = hnametable[idx];

    if(p != NULL) {
      if(p->addr.s_addr == addr)
	break; /* entry found */
    } else if(p == NULL) { /* New entry */
      p = (struct hnamemem*)malloc(sizeof(struct hnamemem));
      memset(p, 0, sizeof(struct hnamemem));
      hnametable[idx] = p;
      
      p->addr.s_addr = addr;

#if defined(MULTITHREADED) && defined(ASYNC_ADDRESS_RESOLUTION)
      p->name = (char*)malloc(maxNameLen);
      sprintf(p->name, "*%s*", intoa(hostIpAddress));
      queueAddress(p);
#else
      addr = ntohl(addr); /* This is necessary due to BIG/LITTLE endian crap */

      hp = (struct hostent*)gethostbyaddr((char*)&addr, 4, AF_INET);

      if (hp && (hp->h_name)) {
	char *dotp = (char*)hp->h_name;    

	if(domainName[0] != '\0') {
	  int len = strlen(dotp)-strlen(domainName);
	  
	  if((len > 0) && (strcmp(&dotp[len], domainName) == 0))
	    dotp[len-1] = '\0';
	}

	p->name = savestr(dotp);

	/* Add '...' at the end of each host name if the name
	 * is longer than maxNameLen chars 
	 * (maxNameLen is the max host name length)
	 * Benedikt Heinen <Benedikt.Heinen@infrasys.ascom.ch>
	 */
	if(strlen(p->name) > maxNameLen) { 
	  p->name[maxNameLen] = '\0';
	  p->name[maxNameLen-1] = '.';
	  p->name[maxNameLen-2] = '.';
	  p->name[maxNameLen-3] = '.';
	}
      } else {
	p->name = savestr(intoa(hostIpAddress));
	/* printf("Unable to find IP address for %s\n", p->name); */
      }
#endif
      break;
    }
    
    idx = (idx+1)%HASHNAMESIZE;
  }

  /* printf("Found %s\n", p->name); */

  return(p->name);
}

/* ******************************* */

int checkKeyPressed() {
  int theChar = getch();
  unsigned char in_char;
  int rc = 1; 

  if((theChar != EOF) 
     && ((in_char = (unsigned char)theChar)) != 255) {
  
    /*
      char buf[32];

      sprintf(buf, ">%c-%d<", in_char, in_char);
      mvprintw(0, 40, buf); refresh();
      sleep(2);
    */
    if ((in_char == 'q') || (in_char == 'Q'))
      cleanup(-1);
    else if ((in_char == 'l') || (in_char == 'L'))
      localAddrFlag = !localAddrFlag; /* toggle host list */
    else if ((in_char == 'd') || (in_char == 'D'))
      idleFlag = !idleFlag; /* toggle idle hosts */
    else if ((in_char == 'p') || (in_char == 'P'))
      percentMode = (percentMode+1)%3; /* toggle traffic percentage */
    else if ((in_char == 'r') || (in_char == 'R'))
      resetStats();
    else if ((in_char == 'n') || (in_char == 'N'))
      numericFlag = ((numericFlag+1) % 4); /* toggle numeric <-> sym ... address format */
    else if ((in_char == 't') || (in_char == 'T')) 
      sortSendMode = !sortSendMode;
    else if ((in_char == 'y') || (in_char == 'y')) 
      columnSort = (columnSort+1)%4;
    else if (in_char == ' ') {
      int numScreens = ((numIpProtosToMonitor+2 /* avoid round problems */)/3);
      if((numIpProtosToMonitor%3) != 0)
	numScreens++;
      numScreens += 3; /* Protocol Screens */
      screenNumber = (screenNumber+1)%numScreens;
    } else
      printHelp();
  } else
    rc = 0;

  if(rc) 
    printHeader(0);

  return(rc);
}

/* ******************************* */

void updateThpt() {
  time_t timeDiff, totalTime;
  int idx;
  HostTraffic *el;

  timeDiff = actTime-lastThptUpdate; 

  /* printf("%u %u %u\n", timeDiff, throughput, ethernetBytes); */

  if(timeDiff > 5 /* secs */) {
    throughput = ethernetBytes-throughput;

    lastThptUpdate = timeDiff;
    actualThpt = (float)throughput/(float)lastThptUpdate;    
    lastThptUpdate = actTime;
    
    if(actualThpt > peakThroughput)
      peakThroughput = actualThpt;

    throughput = ethernetBytes;

    if((timeDiff = actTime-lastMinThptUpdate) > 60 /* 1 minute */) {
      lastMinEthernetBytes = ethernetBytes - lastMinEthernetBytes;
      lastMinThptUpdate = timeDiff;    
      lastMinThpt = (float)lastMinEthernetBytes/(float)lastMinThptUpdate;      
      lastMinEthernetBytes = ethernetBytes;
      lastMinThptUpdate = actTime;
    }
    
    if((timeDiff = actTime-lastFiveMinsThptUpdate) > 300 /* 5 minutes */) {
      lastFiveMinsEthernetBytes = ethernetBytes - lastFiveMinsEthernetBytes;
      lastFiveMinsThptUpdate = timeDiff;
      lastFiveMinsThpt = (float)lastFiveMinsEthernetBytes/(float)lastFiveMinsThptUpdate;      
      lastFiveMinsEthernetBytes = ethernetBytes;
      lastFiveMinsThptUpdate = actTime;
    }    

    /* *********************************************** */        
    totalTime = actTime-initialSniffTime;   
    
    for(idx=0; idx<HASHNAMESIZE; idx++) {
      if((el = hash_hostTraffic[idx]) != NULL) {       
	el->actualRcvdThpt  = (float)(el->bytesReceived-el->lastBytesReceived)/timeDiff;
	el->actualSentThpt  = (float)(el->bytesSent-el->lastBytesSent)/timeDiff;
	el->averageRcvdThpt = ((float)el->bytesReceived)/totalTime;
	el->averageSentThpt = ((float)el->bytesSent)/totalTime;
	hash_hostTraffic[idx]->lastBytesSent     = hash_hostTraffic[idx]->bytesSent;
	hash_hostTraffic[idx]->lastBytesReceived = hash_hostTraffic[idx]->bytesReceived;
      }
    }
  }
}

/* ************************************ */

char* savestr(const char *str)
{
  u_int size;
  char *p;
  static char *strptr = NULL;
  static u_int strsize = 0;

  size = strlen(str) + 1;
  if (size > strsize) {
    strsize = 1024;
    if (strsize < size)
      strsize = size;
    strptr = (char *)malloc(strsize);
    if (strptr == NULL) {
      fprintf(stderr, "savestr: malloc\n");
      exit(1);
    }
  }
  (void)strcpy(strptr, str);
  p = strptr;
  strptr += size;
  strsize -= size;
  return (p);
}

/* ************************************ */

static struct enamemem* lookup_emem(const u_char *ep)
{
  u_int i, j, k;
  struct enamemem *tp;

  k = (ep[0] << 8) | ep[1];
  j = (ep[2] << 8) | ep[3];
  i = (ep[4] << 8) | ep[5];

  tp = &enametable[(i ^ j) & (HASHNAMESIZE-1)];
  while (tp->e_nxt)
    if (tp->e_addr0 == i &&
	tp->e_addr1 == j &&
	tp->e_addr2 == k)
      return tp;
    else
      tp = tp->e_nxt;
  tp->e_addr0 = i;
  tp->e_addr1 = j;
  tp->e_addr2 = k;
  tp->e_nxt = (struct enamemem *)calloc(1, sizeof(*tp));
  if (tp->e_nxt == NULL)
    error("lookup_emem: calloc");

  return tp;
}

/* ************************************ */

char* etheraddr_string(const u_char *ep)
{
  u_int i, j;
  char *cp;
  struct enamemem *tp;
  char buf[sizeof("00:00:00:00:00:00")];

  tp = lookup_emem(ep);

  if (tp->e_name)
    return (tp->e_name);

  cp = buf;

  if ((j = *ep >> 4) != 0)
    *cp++ = hex[j];
  else
    *cp++ = '0';

  *cp++ = hex[*ep++ & 0xf];

  for(i = 5; (int)--i >= 0;) {
    *cp++ = ':';
    if ((j = *ep >> 4) != 0)
      *cp++ = hex[j];
    else
      *cp++ = '0';

    *cp++ = hex[*ep++ & 0xf];
  }

  *cp = '\0';
  tp->e_name = savestr(buf);
  return (tp->e_name);
}

/* ************************************ */

/* Return a zero'ed hnamemem struct and cuts down on calloc() overhead */
struct hnamemem * newhnamemem(void)
{
  struct hnamemem *p;
  static struct hnamemem *ptr = NULL;
  static u_int num = 0;

  if (num  <= 0) {
    num = 64;
    ptr = (struct hnamemem *)calloc(num, sizeof (*ptr));
    if (ptr == NULL)
      error("newhnamemem: calloc");
  }
  --num;
  p = ptr++;
  return (p);
}

/* ************************************ */

/*
 * A faster replacement for inet_ntoa().
 */
char* intoa(struct in_addr addr)
{
  char *cp, *retStr;
  u_int byte;
  int n;
  static char buf[sizeof(".xxx.xxx.xxx.xxx")];
  
  if(addr.s_addr == 0x0)
    return("0.0.0.0");
  else if(addr.s_addr == INADDR_BROADCAST) {
    if(!webMode)
      return("<broadcast>");
    else
      return("&lt;broadcast&gt;");   
  }  else if(isBroadcastAddress(&addr)) {
    if(!webMode)
      return("<subnet broadcast>");
    else
      return("&lt;subnet&nbsp;broadcast&gt;");
  }

  cp = &buf[sizeof buf];
  *--cp = '\0';

  n = 4;
  do {
    byte = addr.s_addr & 0xff;
    *--cp = byte % 10 + '0';
    byte /= 10;
    if (byte > 0) {
      *--cp = byte % 10 + '0';
      byte /= 10;
      if (byte > 0)
	*--cp = byte + '0';
    }
    *--cp = '.';
    addr.s_addr >>= 8;
  } while (--n > 0);

  /* Convert the string to lowercase */
  retStr = (char*)(cp+1);
  for(n=0; retStr[n] != '\0'; n++)
    retStr[n] = tolower(retStr[n]);

  return(cp+1);
}

/* ************************************ */

char* getNamedPort(int port) {
  static char outStr[2][8];
  static short portBufIdx=0;
  char* svcName;

  portBufIdx = (portBufIdx+1)%2;

  svcName = getPortByNum(port, IPPROTO_TCP);
  if(svcName == NULL)
    svcName = getPortByNum(port, IPPROTO_UDP);

  if(svcName == NULL)
    sprintf(outStr[portBufIdx], "%d", port);
  else
    strcpy(outStr[portBufIdx], svcName);

  return(outStr[portBufIdx]);
}

/* ************************************ */

void updateHostSessionsList(HostTraffic *theHost,
			    u_short port,
			    HostTraffic *remotePeer,
			    IPSession *theSession,
			    u_short sessionType,
			    u_char initiator,
			    int role)
{
  /* This is a known port hence we're interested in */
  struct ipGlobalSession *scanner;
  int i;
  
  if(theHost->broadcastHost) return; /* We could't care less of junk traffic */

  switch(sessionType) {
  case IPPROTO_TCP:
    scanner = theHost->tcpSessionList; 
    break;
  case IPPROTO_UDP:
    scanner = theHost->udpSessionList; 
    break;
  }
      
  while(scanner != NULL) {
    if(scanner->port == port) {
      if (scanner->initiator == role)
	break;
    }
      
    scanner = (IpGlobalSession*)(scanner->next);
  }
    
  if(scanner == NULL) {
    scanner = (IpGlobalSession *)calloc(1, sizeof(IpGlobalSession));
    scanner->port = port;
    scanner->initiator = role;
    scanner->firstSeen = actTime;
    /* Add the session to the session list */      
    switch(sessionType) {
    case IPPROTO_TCP:
      scanner->next = (struct ipGlobalSession*)(theHost->tcpSessionList);	
      theHost->tcpSessionList = scanner; /* list head */    
      break;
    case IPPROTO_UDP:
      scanner->next = (struct ipGlobalSession*)(theHost->udpSessionList);	
      theHost->udpSessionList = scanner; /* list head */
      break;
    }
  }

  scanner->lastSeen = actTime;
  scanner->sessionCounter++;

#ifdef DEBUG
  if(logFlag)
    printSession(theSession, sessionType, scanner->sessionCounter);
#endif

  for(i=0; i<MAX_NUM_SESSION_PEERS; i++)
    if((scanner->peers[i] == NULL)
       || (scanner->peers[i] == remotePeer))
      break;
  
  if(scanner->peers[i] != remotePeer) {
    scanner->peers[scanner->lastPeer] = remotePeer;
    scanner->lastPeer = (scanner->lastPeer+1) % MAX_NUM_SESSION_PEERS;
  }

  switch(sessionType) {
  case IPPROTO_TCP:
    if(role == SERVER_ROLE) {
      scanner->bytesSent += theSession->bytesSent;
      scanner->bytesReceived += theSession->bytesReceived;
    } else {
      scanner->bytesSent += theSession->bytesReceived;
      scanner->bytesReceived += theSession->bytesSent;
    }
    break;
  case IPPROTO_UDP:
    scanner->bytesSent += theSession->bytesSent;
    scanner->bytesReceived += theSession->bytesReceived;
    break;
  }
}

/* ************************************ */

static void purgeIdleHostSessions(HostTraffic *host, 
				  IpGlobalSession *sessionScanner) 
{
  int i;
  struct hostTraffic **scanner;

  while(sessionScanner != NULL) {
    scanner = sessionScanner->peers;
    if(scanner != NULL) {
      for(i=0; i<MAX_NUM_SESSION_PEERS; i++)
	if(scanner[i] == host)
	  scanner[i] = NULL;
    }

    sessionScanner = sessionScanner->next;
  }
  
}

/* ************************************ */

static void freeHostInfo(HostTraffic *host) {
  int i, idx, j;

  free(host->protoIPTrafficInfos);

  for(i=0; i<2; i++) {
    IpGlobalSession *nextElement, *element;
    
    if(i == 0)
      element = host->tcpSessionList;
    else
      element = host->udpSessionList;
    
    while(element != NULL) {
      nextElement = element->next;
      /* 
	 The 'peers' field shouldn't be a problem because an idle host
	 isn't supposed to have any session 
      */
      free(element);
      element = nextElement;
    }
  }

  /* Check whether there are hosts that have the host being purged
     as peer */
  for(idx=0; idx<HASHNAMESIZE; idx++)
   if(hash_hostTraffic[idx] != NULL) { 
     if(hash_hostTraffic[idx]->tcpSessionList != NULL)
       purgeIdleHostSessions(host, hash_hostTraffic[idx]->tcpSessionList);
     
     if(hash_hostTraffic[idx]->udpSessionList != NULL)
       purgeIdleHostSessions(host, hash_hostTraffic[idx]->udpSessionList);	     

     for(j=0; j<MAX_NUM_CONTACTED_PEERS; j++)
       if(hash_hostTraffic[idx]->contactedIpPeers[j] == host)
	 hash_hostTraffic[idx]->contactedIpPeers[j] = NULL;
   }

  free(host);
}

/* ************************************ */
  
void purgeIdleHosts() {
  int idx;
  
  purgeOldFragmentEntries(); /* let's do this too */

  for(idx=0; idx<HASHNAMESIZE; idx++)
    if((hash_hostTraffic[idx] != NULL) 
       && (!hash_hostTraffic[idx]->subnetLocalHost)
       && ((hash_hostTraffic[idx]->lastSeen+IDLE_HOST_PURGE_TIMEOUT) < actTime)) {
      HostTraffic* hostToFree;
      hostToFree = hash_hostTraffic[idx];
      hash_hostTraffic[idx] = NULL;
      freeHostInfo(hostToFree); /* it needs thread synch */
    }
}

/* ************************************ */

void scanTimedoutTCPSessions() {
  int idx;

#ifdef DEBUG
  printf("Called scanTimedoutTCPSessions\n");
#endif

  for(idx=0; idx<HASHNAMESIZE; idx++) {
    if(tcpSession[idx] != NULL) {
#ifdef DEBUG
      printf("Session to scan: %s:%d <--> %s:%d (%lu/%lu)\n",
	     ipaddr2str(tcpSession[idx]->initiator->hostIpAddress), tcpSession[idx]->sport,
	     ipaddr2str(tcpSession[idx]->remotePeer->hostIpAddress), tcpSession[idx]->dport,
	     tcpSession[idx]->lastSeen+TWO_MSL_TIMEOUT,
	     actTime);
#endif      
      if(
	 ((tcpSession[idx]->sessionState == STATE_TIMEOUT)
	    && ((tcpSession[idx]->lastSeen+TWO_MSL_TIMEOUT) < actTime))
	 || /* The branch below allows to flush sessions which have not been
	       terminated properly (we've received just one FIN (not two). It might be
	       that we've lost some packets (hopefully not). */
	 ((tcpSession[idx]->sessionState >= STATE_FIN1_ACK0)
	    && ((tcpSession[idx]->lastSeen+DOUBLE_TWO_MSL_TIMEOUT) < actTime))
	 /* The line below allows to avoid keeping very old sessions that
	    might be still open, but that are probably closed and we've
	    lost some packets */
	 || ((tcpSession[idx]->lastSeen+IDLE_HOST_PURGE_TIMEOUT) < actTime)
	 )
	{
	  /* Session to purge */
#ifdef DEBUG
	  printf("Session completed: %s:%d <--> %s:%d \n",
		 ipaddr2str(tcpSession[idx]->initiator->hostIpAddress), tcpSession[idx]->sport,
		 ipaddr2str(tcpSession[idx]->remotePeer->hostIpAddress), tcpSession[idx]->dport);
	  if(logFlag) printSession(tcpSession[idx], IPPROTO_TCP, 0);
#endif
	 
	  if(tcpSession[idx]->sport < tcpSession[idx]->dport) { /* Server -> Client */
	    if(getPortByNum(tcpSession[idx]->sport, IPPROTO_TCP) != NULL) {
	      updateHostSessionsList(tcpSession[idx]->initiator, tcpSession[idx]->sport, 
				     tcpSession[idx]->remotePeer, tcpSession[idx], 
				     IPPROTO_TCP, SERVER_TO_CLIENT, SERVER_ROLE);
	      updateHostSessionsList(tcpSession[idx]->remotePeer, tcpSession[idx]->sport, 
				     tcpSession[idx]->initiator, tcpSession[idx], 
				     IPPROTO_TCP, CLIENT_FROM_SERVER, CLIENT_ROLE);
	    }
	  } else { /* Client -> Server */
	    if(getPortByNum(tcpSession[idx]->dport, IPPROTO_TCP) != NULL) {
	      updateHostSessionsList(tcpSession[idx]->remotePeer, tcpSession[idx]->dport, 
				     tcpSession[idx]->initiator, tcpSession[idx], 
				     IPPROTO_TCP, SERVER_FROM_CLIENT, SERVER_ROLE);
	      updateHostSessionsList(tcpSession[idx]->initiator, tcpSession[idx]->dport, 
				     tcpSession[idx]->remotePeer, tcpSession[idx], 
				     IPPROTO_TCP, CLIENT_TO_SERVER, CLIENT_ROLE);
	    }
	  }

	  /* 
	   * Having updated the session information, 'theSession'
	   * can now be purged.
	   */
	  free(tcpSession[idx]); /* No inner pointers to free */
	  tcpSession[idx] = NULL;
	  numTcpSessions--;
	}
    }
  } /* end for */

}
  
/* ************************************ */


void handleSession(IPSession *sessions[],
		   u_short *numSessions,
		   HostTraffic *srcHost,
		   u_short sport,
		   HostTraffic *dstHost,
		   u_short dport,
		   u_int length,                   
                   struct tcphdr *tp,
		   u_int tcpDataLength)
{
  u_int idx, initialIdx;
  IPSession *theSession = NULL;
  short flowDirection;
  u_short sessionType, check;
  u_short sessSport, sessDport;
  int j;

  if(srcHost->broadcastHost || dstHost->broadcastHost)
    return;

#ifdef DEBUG
  printf("%8x/%8x %8x\n", 
	 srcHost->hostIpAddress.s_addr,
	 dstHost->hostIpAddress.s_addr,
	 localHostAddress.s_addr);
#endif

#ifdef HAVE_LSOF
  if((srcHost->hostIpAddress.s_addr == localHostAddress.s_addr)
     && (localPorts[sport] != NULL)) {
    for(j=0; j<MAX_NUM_SHARED_PORTS; j++)
      if(localPorts[sport][j] == NULL)
	break;
      else
	localPorts[sport][j]->bytesSent += length;
  }

  if((dstHost->hostIpAddress.s_addr == localHostAddress.s_addr)
     && (localPorts[dport] != NULL)) {
    for(j=0; j<MAX_NUM_SHARED_PORTS; j++)
      if(localPorts[dport][j] == NULL)
	break;
      else
	localPorts[dport][j]->bytesReceived += length;
  }
#endif

  if(tp == NULL)
    sessionType = IPPROTO_UDP;
  else {
    sessSport = ntohs(tp->th_sport);
    sessDport = ntohs(tp->th_dport);   
    sessionType = IPPROTO_TCP;
  }

  /* 
   * The hash key has to be calculated in a specular
   * way: its value has to be the same regardless 
   * of the flow direction.
   *
   * Patch on the line below courtesy of  
   * Paul Chapman <pchapman@fury.bio.dfo.ca>
   */
  initialIdx = idx = (u_int)(((u_long)srcHost+(u_long)dstHost+
			      (u_long)sport+(u_long)dport) % HASHNAMESIZE);

  if(sessionType == IPPROTO_TCP) {
    for(;;) {
      theSession = sessions[idx];
      if(theSession != NULL) { 
	if((theSession->initiator == srcHost) && (theSession->remotePeer == dstHost)
	   && (theSession->sport == sport) && (theSession->dport == dport)) {
	  flowDirection = CLIENT_TO_SERVER;
	  break;
	} else if((theSession->initiator == dstHost) && (theSession->remotePeer == srcHost)
		  && (theSession->sport == dport) && (theSession->dport == sport)) {
	  flowDirection = SERVER_TO_CLIENT;
	  break;
	}
      } else if(theSession == NULL) { 
	/* New Session */

	if(tp->th_flags & TH_FIN) {
	  /* We've received a FIN for a session that
	     we've not seen (ntop started when the 
	     session completed).
	  */
#ifdef DEBUG
	  printf("Received FIN %u for unknown session\n",  ntohl(tp->th_seq)+tcpDataLength);
#endif
	  return; /* Nothing else to do */
	}

	if((*numSessions) > (HASHNAMESIZE/2)) {
	  /* The hash table is getting large: let's replace the oldest session
	     with this one we're allocating */
	  u_int usedIdx;
	
	  for(idx=0; idx<HASHNAMESIZE; idx++) {
	    if(sessions[idx] != NULL) {
	      if(theSession == NULL) {
		theSession = sessions[idx];
		usedIdx = idx;
	      }
	      else if(theSession->lastSeen > sessions[idx]->lastSeen) {
		theSession = sessions[idx];
		usedIdx = idx;
	      }
	    }
	  }

	  sessions[usedIdx] = NULL;
	} else { 
	  /* There's enough space left in the hashtable */
	  theSession = (IPSession*)calloc(1, sizeof(IPSession));
	  (*numSessions)++;
	}

	if(sport < 1024) {
	  srcHost->usedServerPorts[sport]++;
	  dstHost->usedClientPorts[sport]++; 
	}
	if(dport < 1024) {
	  srcHost->usedClientPorts[dport]++; 
	  dstHost->usedServerPorts[dport]++;
	}
	
	while(sessions[initialIdx] != NULL)
	  initialIdx = ((initialIdx+1) % HASHNAMESIZE);

	sessions[initialIdx] = theSession;

	theSession->initiator = srcHost;
	theSession->remotePeer = dstHost;
	theSession->sport = sport;
	theSession->dport = dport;
	theSession->firstSeen = actTime;
	flowDirection = CLIENT_TO_SERVER;

#ifdef DEBUG
	if(logFlag) printSession(theSession, sessionType, 0);
#endif
	break;
      }

      idx = ((idx+1) % HASHNAMESIZE);
    }
  
    theSession->lastSeen = actTime;

    if(flowDirection == CLIENT_TO_SERVER)
      theSession->bytesSent += length;
    else
      theSession->bytesReceived += length;

    /*
     *
     * In this case the session is over hence the list of
     * sessions initiated/received by the hosts can be updated
     *
     */

    if(tp->th_flags & TH_FIN) {
      u_int32_t fin = ntohl(tp->th_seq)+tcpDataLength;

      if(sport < dport) /* Server -> Client */
	check = (fin != theSession->lastSCFin);
      else /* Client -> Server */
	check = (fin != theSession->lastCSFin);

      if(check) {
	/* This is not a duplicated (retransmitted) FIN */
	theSession->finId[theSession->numFin] = fin;
	theSession->numFin++;

      if(sport < dport) /* Server -> Client */
	theSession->lastSCFin = fin;
      else /* Client -> Server */
	theSession->lastCSFin = fin;
	switch(theSession->sessionState) {
	  case STATE_BEGIN:
	    theSession->sessionState = STATE_FIN1_ACK0;
	    break;
	  case STATE_FIN1_ACK0:
	    theSession->sessionState = STATE_FIN1_ACK1;
	    break;
	  case STATE_FIN1_ACK1:
	    theSession->sessionState = STATE_FIN2_ACK1;
	    break;
#ifdef DEBUG
	default:
	  printf("ERROR: unable to handle received FIN (%u) !\n", fin);
#endif
	}
      } else {
#ifdef DEBUG
	printf("Rcvd Duplicated FIN %u\n", fin);
#endif
      }
    } else if((tp->th_flags & TH_ACK) && (theSession->numFin > 0)) { 
      u_int32_t ack = ntohl(tp->th_ack);
      int i;

      if(sport < dport) /* Server -> Client */
	check = (ack != theSession->lastSCAck);
      else /* Client -> Server */
	check = (ack != theSession->lastCSAck);

      if(check) {
	/* This is not a duplicated ACK */

	if(sport < dport) /* Server -> Client */
	  theSession->lastSCAck = ack;
	else /* Client -> Server */
	  theSession->lastCSAck = ack;
	
	for(i=0; i<theSession->numFin; i++) {
	  if((theSession->finId[i]+1) == ack) {
	    theSession->numFinAcked++;
	    theSession->finId[i] = 0;

	    switch(theSession->sessionState) {
	    case STATE_FIN1_ACK0:
	      theSession->sessionState = STATE_FIN1_ACK1;
	      break;
	    case STATE_FIN2_ACK0:
	      theSession->sessionState = STATE_FIN2_ACK1;
	      break;
	    case STATE_FIN2_ACK1:
	      theSession->sessionState = STATE_FIN2_ACK2;
	      break;
#ifdef DEBUG
	    default:
	      printf("ERROR: unable to handle received ACK (%u) !\n", ack);
#endif
	    }	    
	    break;
	  }
	}
      } else {
      }
    }

    theSession->lastFlags = tp->th_flags;

    if((theSession->sessionState == STATE_FIN2_ACK2)
       || ((tp->th_flags & TH_RST) != 0)) /* abortive release */ {
      theSession->sessionState = STATE_TIMEOUT;      
    }
  } else if(sessionType == IPPROTO_UDP) { 
    IPSession tmpSession;

    if(sport < 1024) {
      srcHost->usedServerPorts[sport]++;
      dstHost->usedClientPorts[sport]++; 
    }

    if(dport < 1024) {
      srcHost->usedClientPorts[dport]++; 
      dstHost->usedServerPorts[dport]++;
    }
    
    tmpSession.lastSeen = actTime;
    tmpSession.initiator = srcHost, tmpSession.remotePeer = dstHost;
    tmpSession.bytesSent = length, tmpSession.bytesReceived = 0;
    tmpSession.sport = sport, tmpSession.dport = dport;

#ifdef DEBUG
    if(logFlag) printSession(&tmpSession, sessionType, 0);
#endif

   if(getPortByNum(sport, sessionType) != NULL) {
     updateHostSessionsList(srcHost, sport, dstHost, &tmpSession, 
			    sessionType, SERVER_TO_CLIENT, SERVER_ROLE);
     tmpSession.bytesSent = 0, tmpSession.bytesReceived = length;
     updateHostSessionsList(dstHost, sport, srcHost, &tmpSession, 
			    sessionType, CLIENT_FROM_SERVER, CLIENT_ROLE);
   }
   
   if(getPortByNum(dport, sessionType) != NULL) {
     updateHostSessionsList(srcHost, dport, dstHost, &tmpSession, 
			    sessionType, CLIENT_TO_SERVER, CLIENT_ROLE);
     tmpSession.bytesSent = 0, tmpSession.bytesReceived = length;
     updateHostSessionsList(dstHost, dport, srcHost, &tmpSession, 
			    sessionType, SERVER_FROM_CLIENT, SERVER_ROLE);
   } 
  }  
}

/* ************************************ */

void handleTCPSession(HostTraffic *srcHost,
		      u_short sport,
		      HostTraffic *dstHost,
		      u_short dport,
		      u_int length, 
		      struct tcphdr *tp,
		      u_int tcpDataLength) {

  if((tp->th_flags & TH_SYN) == 0) {
    /* When we receive SYN it means that the client is trying to
       open a session with the server hence the session is NOT YET
       open. That's why we don't count this as a session in this
       case.
    */
    handleSession(tcpSession, &numTcpSessions, 
		  srcHost, sport, 
		  dstHost, dport, 
		  length, tp, tcpDataLength);
  }
}

/* ************************************ */

void handleUDPSession(HostTraffic *srcHost,
		      u_short sport,
		      HostTraffic *dstHost,
		      u_short dport,
		      u_int length) {
  handleSession(udpSession, &numUdpSessions, srcHost, sport,
		dstHost, dport, length, NULL, 0);
}

/* ************************************ */

int handleIP(u_short port,   
	     HostTraffic *srcHost,
	     HostTraffic *dstHost,
	     u_int length) {
  
  int idx = mapGlobalToLocalIdx(port);
  
  if(idx != -1) {
    if(srcHost->subnetPseudoLocalHost == LOCAL_HOST) {
      if(dstHost->subnetPseudoLocalHost == LOCAL_HOST) {
	if(!srcHost->broadcastHost) srcHost->protoIPTrafficInfos[idx].sentLocally += length;
	if(!dstHost->broadcastHost) dstHost->protoIPTrafficInfos[idx].receivedLocally += length;
	ipProtoStats[idx].local += length;
      } else {
	if(!srcHost->broadcastHost) srcHost->protoIPTrafficInfos[idx].sentRemotely += length;
	if(!dstHost->broadcastHost) dstHost->protoIPTrafficInfos[idx].receivedLocally += length;
	ipProtoStats[idx].local2remote += length;
      }
    } else {
      /* srcHost is remote */
      if(dstHost->subnetPseudoLocalHost == LOCAL_HOST) {
	if(!srcHost->broadcastHost) srcHost->protoIPTrafficInfos[idx].sentLocally += length;
	if(!dstHost->broadcastHost) dstHost->protoIPTrafficInfos[idx].receivedFromRemote += length;
	ipProtoStats[idx].remote2local += length;
      } else {
	if(!srcHost->broadcastHost) srcHost->protoIPTrafficInfos[idx].sentRemotely += length;
	if(!dstHost->broadcastHost) dstHost->protoIPTrafficInfos[idx].receivedFromRemote += length;
	ipProtoStats[idx].remote += length;
      }
    }    
  }

  return(idx);
}

/* ************************************ */

void addContactedIPPeers(HostTraffic *peer_a, 
			 HostTraffic *peer_b) {
  short i;

  if(peer_a->broadcastHost || peer_b->broadcastHost)
    return;

  for(i=0; i<MAX_NUM_CONTACTED_PEERS; i++)
    if(peer_a->contactedIpPeers[i] == peer_b)
      return;

  peer_a->contactedIpPeers[peer_a->contactedIpPeersIdx] = peer_b;
  peer_a->contactedIpPeersIdx = (peer_a->contactedIpPeersIdx+1) % MAX_NUM_CONTACTED_PEERS;
}

/* ************************************ */

void addFragmentInfo(HostTraffic *srcHost, 
		     HostTraffic *dstHost,
		     u_short sport, 
		     u_short dport, 
		     u_int fragmentId) {
  IpFragment *fragment;
  u_int idx, startIdx;

  startIdx = idx = ((u_int)((unsigned long)srcHost+(unsigned long)dstHost)*fragmentId) % HASHNAMESIZE;

#ifdef DEBUG
  printf("Added fragmentId %u\n", fragmentId);
#endif

  while(1) {
    fragment = fragmentTable[idx];

    if(fragment == NULL) {
      fragment = (IpFragment*)calloc(1, sizeof(IpFragment));
      fragment->src = srcHost;
      fragment->dest = dstHost;
      fragment->fragmentId = fragmentId;
      fragment->sport = sport;
      fragment->dport = dport;
      fragment->firstSeen = actTime;
      fragmentTable[idx] = fragment;
      return;
    }

    idx = (idx+1) % HASHNAMESIZE;

    if(startIdx == idx) {
      printf("ERROR: the fragment table is full!\n");
      break;
    }
  }
}


/* ************************************ */

void getFragmentInfo(HostTraffic *srcHost, 
		     HostTraffic *dstHost,
		     u_short *sport, 
		     u_short *dport, 
		     u_int fragmentId,
		     u_short fragmentAction) {

  IpFragment *fragment;
  u_int idx, startIdx;

  startIdx = idx = ((u_int)((unsigned long)srcHost+(unsigned long)dstHost)*fragmentId) % HASHNAMESIZE;

  while(1) {
    fragment = fragmentTable[idx];

    if((fragment != NULL) 
       && (fragment->src == srcHost)
       && (fragment->dest == dstHost)
       && (fragment->fragmentId == fragmentId)) {
      (*sport) = fragment->sport;
      (*dport) = fragment->dport;

#ifdef DEBUG
      printf("Found fragmentId %u\n", fragmentId);
#endif
      if(fragmentAction == DELETE_FRAGMENT) {
	free(fragment);
	fragmentTable[idx] = NULL;
#ifdef DEBUG
	printf("Removed fragmentId %u entry\n", fragmentId);
#endif
      }

      break;
    }

    idx = (idx+1) % HASHNAMESIZE;

    if(startIdx == idx) {
#ifdef DEBUG
      printf("ERROR: unable to find the expected fragment %u!\n", 
	     fragmentId);
#endif
      break;
    }
  }

}

/* ************************************ */

void purgeOldFragmentEntries() {
  int i;

  for(i=0; i<HASHNAMESIZE; i++) {
    if((fragmentTable[i] != NULL)
       && ((fragmentTable[i]->firstSeen+DOUBLE_TWO_MSL_TIMEOUT) > actTime)) {
	free(fragmentTable[i]);
	fragmentTable[i] = NULL;
    }
  }
}

/* ************************************ */

void processIpPkt(const u_char *bp, u_int length,
		  u_char *ether_src, u_char *ether_dst)
{
  u_short sport, dport;
  struct ip ip;
  struct tcphdr tp;
  struct udphdr up;
  u_int hlen, tcpDataLength, off;
  char *proto;
  HostTraffic *srcHost, *dstHost;
  register u_char saveflags;
  u_char flags = '.';

  /* Need to copy this over in case bp isn't properly aligned.
   * This occurs on SunOS 4.x at least.                
   * Paul D. Smith <psmith@baynetworks.com>
   */
  memcpy(&ip, bp, sizeof(struct ip));
  hlen = (u_int)ip.ip_hl * 4;

  NTOHL(ip.ip_src.s_addr); srcHost = getHostInfo(&ip.ip_src, ether_src);
  NTOHL(ip.ip_dst.s_addr); dstHost = getHostInfo(&ip.ip_dst, ether_dst);  

  addContactedIPPeers(srcHost, dstHost);
  addContactedIPPeers(dstHost, srcHost);

  updatePacketCount(srcHost, dstHost, length);
  updateTrafficMatrix(srcHost, dstHost, length);

  if(srcHost->subnetPseudoLocalHost == LOCAL_HOST) {
    if(dstHost->subnetPseudoLocalHost == LOCAL_HOST) {
      srcHost->bytesSentLocally += length;
      dstHost->bytesReceivedLocally += length;
    } else {
      srcHost->bytesSentRemotely += length;
      dstHost->bytesReceivedLocally += length;
    }
  } else {
    /* srcHost is remote */
    if(dstHost->subnetPseudoLocalHost == LOCAL_HOST) {
      srcHost->bytesSentLocally += length;
      dstHost->bytesReceivedFromRemote += length;
    } else {
      srcHost->bytesSentRemotely += length;
      dstHost->bytesReceivedFromRemote += length;
    }
  }    

  ipBytes += length;
  off = ntohs(ip.ip_off);

  switch(ip.ip_p) {
  case IPPROTO_TCP:
    proto = "TCP";
    tcpBytes += length;
    memcpy(&tp, bp+hlen, sizeof(struct tcphdr));

    tcpDataLength = ntohs(ip.ip_len) - hlen - (tp.th_off * 4);

    sport = ntohs(tp.th_sport);
    dport = ntohs(tp.th_dport);
       
    if (off & 0x3fff) { /* Handle fragmented packets */
      u_int fragmentOffset = (off & 0x1fff) * 8;

      if(fragmentOffset == 0) /* First fragment */
	addFragmentInfo(srcHost, dstHost, sport, dport, ntohs(ip.ip_id));
      else {
	if(!(off & IP_MF)) {
	  /* This is the last fragment */
	  getFragmentInfo(srcHost, dstHost, &sport, 
			  &dport, ntohs(ip.ip_id), DELETE_FRAGMENT);
	} else {
	  /* Intermediate fragment */
	  getFragmentInfo(srcHost, dstHost, &sport, 
			  &dport, ntohs(ip.ip_id), KEEP_FRAGMENT);
	}
      }

      tcpDataLength = ntohs(ip.ip_len)-hlen;
    }

    sport = ntohs(tp.th_sport);
    dport = ntohs(tp.th_dport);
    if ((saveflags = tp.th_flags) & (TH_SYN|TH_FIN|TH_RST|TH_PUSH)) {
      if (saveflags & TH_SYN)
	flags = 'S';
      if (saveflags & TH_FIN)
	flags = 'F';
      if (saveflags & TH_RST)
	flags = 'R';
      if (saveflags & TH_PUSH)
	flags = 'P';
    } 

    if(srcHost->subnetPseudoLocalHost == LOCAL_HOST) {
      if(dstHost->subnetPseudoLocalHost == LOCAL_HOST) {
	srcHost->tcpSentLocally += length;
	dstHost->tcpReceivedLocally += length;
	tcpGlobalTrafficStats.local += length;
      } else {
	srcHost->tcpSentRemotely += length;
	dstHost->tcpReceivedLocally += length;
	tcpGlobalTrafficStats.local2remote += length;
      }
    } else {
      /* srcHost is remote */
      if(dstHost->subnetPseudoLocalHost == LOCAL_HOST) {
	srcHost->tcpSentLocally += length;
	dstHost->tcpReceivedFromRemote += length;
	tcpGlobalTrafficStats.remote2local += length;
      } else {
	srcHost->tcpSentRemotely += length;
	dstHost->tcpReceivedFromRemote += length;
	tcpGlobalTrafficStats.remote += length;
      }
    }    
    
    if(handleIP(dport, srcHost, dstHost, length) == -1)
      handleIP(sport, srcHost, dstHost, length);

    handleTCPSession(srcHost, sport, dstHost, 
		     dport, length, &tp, tcpDataLength);
    break;

  case IPPROTO_UDP:
    proto = "UDP";
    udpBytes += length;
    memcpy(&up, bp+hlen, sizeof(struct udphdr));
    /* print TCP packet useful for debugging */

    sport = ntohs(up.uh_sport);
    dport = ntohs(up.uh_dport);

    if (off & 0x3fff) { /* Handle fragmented packets */
      u_int fragmentOffset = (off & 0x1fff) * 8;

      if(fragmentOffset == 0) /* First fragment */
	addFragmentInfo(srcHost, dstHost, sport, dport, ntohs(ip.ip_id));
      else {
	if(!(off & IP_MF)) {
	  /* This is the last fragment */
	  getFragmentInfo(srcHost, dstHost, &sport, 
			  &dport, ntohs(ip.ip_id), DELETE_FRAGMENT);
	} else {
	  /* Intermediate fragment */
	  getFragmentInfo(srcHost, dstHost, &sport, 
			  &dport, ntohs(ip.ip_id), KEEP_FRAGMENT);
	}
      }

      tcpDataLength = ntohs(ip.ip_len)-hlen;
    }

    if(srcHost->subnetPseudoLocalHost == LOCAL_HOST) {
      if(dstHost->subnetPseudoLocalHost == LOCAL_HOST) {
	srcHost->udpSentLocally += length;
	dstHost->udpReceivedLocally += length;
	udpGlobalTrafficStats.local += length;
      } else {
	srcHost->udpSentRemotely += length;
	dstHost->udpReceivedLocally += length;
	udpGlobalTrafficStats.local2remote += length;
      }
    } else {
      /* srcHost is remote */
      if(dstHost->subnetPseudoLocalHost == LOCAL_HOST) {
	srcHost->udpSentLocally += length;
	dstHost->udpReceivedFromRemote += length;
	udpGlobalTrafficStats.remote2local += length;
      } else {
	srcHost->udpSentRemotely += length;
	dstHost->udpReceivedFromRemote += length;
	udpGlobalTrafficStats.remote += length;
      }
    }        

    if(handleIP(dport, srcHost, dstHost, length) == -1)
      handleIP(sport, srcHost, dstHost, length);

    handleUDPSession(srcHost, sport, dstHost, dport, length);
    break;

  case IPPROTO_ICMP:
    proto = "ICMP";
    icmpBytes += length;
    srcHost->icmpSent += length;
    dstHost->icmpReceived += length;
    break;

  case IPPROTO_OSPF:
    proto = "OSPF";
    ospfBytes += length;
    srcHost->ospfSent += length;
    dstHost->ospfReceived += length;
    break;

  case IPPROTO_IGMP:
    proto = "IGMP";
    igmpBytes += length;
    srcHost->igmpSent += length;
    dstHost->igmpReceived += length;
    break;

  default:
    proto = "IP (Other)";
    otherIpBytes += length;
    sport = dport = -1;
    srcHost->otherSent += length;
    dstHost->otherReceived += length;
    break;
  }
}

/* ************************************ */


char* llcsap_string(u_char sap)
{
  register char *cp;
  static char buf[sizeof("sap 00")];

  cp = buf;
  (void)strcpy(cp, "sap ");
  cp += strlen(cp);
  *cp++ = hex[sap >> 4 & 0xf];
  *cp++ = hex[sap & 0xf];
  *cp++ = '\0';

  /* printf("%s\n", buf); */
  return(buf);
}

/* ************************************ */

void updatePacketCount(HostTraffic *srcHost, HostTraffic *dstHost, 
		       TrafficCounter length) {
  srcHost->pktSent++;

  if(isBroadcastAddress(&(dstHost->hostIpAddress))) {
    srcHost->pktBroadcastSent++;
    broadcastPkts++;
  } else if(isMulticastAddress(&(dstHost->hostIpAddress))) {
#ifdef DEBUG
    printf("%s -> %s\n", srcHost->hostSymIpAddress, dstHost->hostSymIpAddress);
#endif
    srcHost->pktMulticastSent++;
    multicastPkts++;
  }

  srcHost->bytesSent += length;
  dstHost->bytesReceived += length;

  dstHost->pktReceived++;

  if(length < 64) rcvdPktStats.upTo64++;
  else if(length < 128) rcvdPktStats.upTo128++;
  else if(length < 256) rcvdPktStats.upTo256++;
  else if(length < 512) rcvdPktStats.upTo512++;
  else if(length < 1024) rcvdPktStats.upTo1024++;
  else if(length < 1518) rcvdPktStats.upTo1518++;
  else rcvdPktStats.above1518++;

  if((rcvdPktStats.shortest == 0)
     || (rcvdPktStats.shortest > length))
    rcvdPktStats.shortest = length;

  if(rcvdPktStats.longest < length)
    rcvdPktStats.longest = length;
}

/* ************************************ */

void updateTrafficMatrix(HostTraffic *srcHost, HostTraffic *dstHost, TrafficCounter length) {
  int a, b;

  if((srcHost->hostIpAddress.s_addr == INADDR_BROADCAST)
     || (!isPseudoLocalAddress(&(srcHost->hostIpAddress)))
     || (dstHost->hostIpAddress.s_addr == INADDR_BROADCAST)
     || (!isPseudoLocalAddress(&(dstHost->hostIpAddress))))
    return;
  
  a = (TrafficCounter)(srcHost->hostIpAddress.s_addr) & 0x000000FF;
  b = (TrafficCounter)(dstHost->hostIpAddress.s_addr) & 0x000000FF;

  ipTrafficMatrixHosts[a] = srcHost, ipTrafficMatrixHosts[b] = dstHost;
  ipTrafficMatrix[a][b].bytesSent += length, ipTrafficMatrix[b][a].bytesReceived += length;
}

/* ************************************ */

#ifdef MULTITHREADED

void queuePacket(u_char *unused, 
		 const struct pcap_pkthdr *h,
		 const u_char *p) {

  /****************************
    - If the queue is full then wait until a slot is freed
    - If the queue is getting full then periodically wait
      until a slot is freed
  *****************************/

  if(packetQueueLen >= PACKET_QUEUE_LENGTH) {
    printf("Dropping packet!!! [packet queue=%d/max=%d]\n", 
	   packetQueueLen, maxPacketQueueLen);
    droppedPackets++;
#ifdef HAVE_SCHED_H
    sched_yield(); /* Allow other threads (dequeue) to run */
#endif
  } else {  
    memcpy(&packetQueue[packetQueueHead].h, h, sizeof(struct pcap_pkthdr));
    memcpy(packetQueue[packetQueueHead].p, p, DEFAULT_SNAPLEN);
    packetQueueHead = (packetQueueHead+1) % PACKET_QUEUE_LENGTH;
    accessMutex(&packetQueueMutex);
    packetQueueLen++;
    if(packetQueueLen > maxPacketQueueLen)
      maxPacketQueueLen = packetQueueLen;
    releaseMutex(&packetQueueMutex);
#ifdef DEBUG
    printf("Queued packet... [packet queue=%d/max=%d]\n", 
	   packetQueueLen, maxPacketQueueLen);
#endif

#ifdef DEBUG_THREADS
    printf("+ [packet queue=%d/max=%d]\n", packetQueueLen, maxPacketQueueLen);
#endif

#ifdef USE_SEMAPHORES
    incrementSem(&queueSem);
#else
    signalCondvar(&queueCondvar);
#endif
#ifdef HAVE_SCHED_H
    sched_yield(); /* Allow other threads (dequeue) to run */
#endif
  }
}

/* ************************************ */

void* dequeuePacket(void* notUsed) {
  PacketInformation pktInfo;

  while(1) {
#ifdef DEBUG
    printf("Waiting for packet...\n");
#endif

#ifdef USE_SEMAPHORES
    waitSem(&queueSem);
#else
    waitCondvar(&queueCondvar);
#endif
#ifdef DEBUG
    printf("Got packet...\n");
#endif
    memcpy(&pktInfo.h, &packetQueue[packetQueueTail].h, 
	   sizeof(struct pcap_pkthdr));
    memcpy(pktInfo.p, packetQueue[packetQueueTail].p, DEFAULT_SNAPLEN);
    packetQueueTail = (packetQueueTail+1) % PACKET_QUEUE_LENGTH;
    accessMutex(&packetQueueMutex);
    packetQueueLen--;
    releaseMutex(&packetQueueMutex);
#ifdef DEBUG_THREADS
    printf("- [packet queue=%d/max=%d]\n", packetQueueLen, maxPacketQueueLen);
#endif

#ifdef DEBUG
    printf("Processing packet... [packet queue=%d/max=%d]\n", 
	   packetQueueLen, maxPacketQueueLen);
#endif
    pbuf_process(NULL, &pktInfo.h, pktInfo.p);
  }

  return(NULL); /* NOTREACHED */
}

/* ************************************ */

#if defined(MULTITHREADED) && defined(ASYNC_ADDRESS_RESOLUTION)

void queueAddress(struct hnamemem* elem) {
  /****************************
    - If the queue is full then wait until a slot is freed
    - If the queue is getting full then periodically wait
      until a slot is freed
  *****************************/
  if(addressQueueLen >= ADDRESS_QUEUE_LENGTH) {
    printf("Dropping address!!! [addr queue=%d/max=%d]\n", 
	   addressQueueLen, maxAddressQueueLen);
    droppedAddresses++;    
#ifdef HAVE_SCHED_H
    sched_yield(); /* Allow other threads (dequeue) to run */
#endif
  } else {  
    addressQueue[addressQueueHead] = elem;
    addressQueueHead = (addressQueueHead+1) % ADDRESS_QUEUE_LENGTH;
    accessMutex(&addressQueueMutex);
    addressQueueLen++;
    if(addressQueueLen > maxAddressQueueLen)
      maxAddressQueueLen = addressQueueLen;
    releaseMutex(&addressQueueMutex);
#ifdef DEBUG
    printf("Queued address... [addr queue=%d/max=%d]\n", 
	   addressQueueLen, maxAddressQueueLen);
#endif

#ifdef DEBUG_THREADS
    printf("+ [addr queue=%d/max=%d]\n", addressQueueLen, maxAddressQueueLen);
#endif

#ifdef USE_SEMAPHORES
    incrementSem(&queueAddressSem);
#else
    signalCondvar(&queueAddressCondvar);
#endif
  }
}

/* ************************************ */

void* dequeueAddress(void* notUsed) {
  struct hnamemem *elem;
  int addr;
  char* res;
  struct hostent *hp = NULL;

  while(1) {
#ifdef DEBUG
    printf("Waiting for address...\n");
#endif

#ifdef USE_SEMAPHORES
    waitSem(&queueAddressSem);
#else
    waitCondvar(&queueAddressCondvar);
#endif

    elem = addressQueue[addressQueueTail];
    addressQueueTail = (addressQueueTail+1) % ADDRESS_QUEUE_LENGTH;
    accessMutex(&addressQueueMutex);
    addressQueueLen--;
    releaseMutex(&addressQueueMutex);
#ifdef DEBUG_THREADS
    printf("- [addr queue=%d/max=%d]\n", addressQueueLen, maxAddressQueueLen);
#endif

#ifdef DEBUG
      printf("Processing address... [addr queue=%d/max=%d]\n", 
	     addressQueueLen, maxAddressQueueLen);
#endif

      addr = ntohl(elem->addr.s_addr); /* This is necessary due to BIG/LITTLE endian crap */
      hp = (struct hostent*)gethostbyaddr((char*)&addr, 4, AF_INET);

      if (hp && (hp->h_name)) {
	char *dotp = (char*)hp->h_name;    

	if(domainName[0] != '\0') {
	  int len = strlen(dotp)-strlen(domainName);
	  
	  if((len > 0) && (strcmp(&dotp[len], domainName) == 0))
	    dotp[len-1] = '\0';
	}
	res = dotp;
      } else
	res = intoa(elem->addr);

      if(strlen(res) > maxNameLen) {
	strncpy(elem->name, res, maxNameLen-3);
	elem->name[maxNameLen] = '\0';
	elem->name[maxNameLen-1] = '.';
	elem->name[maxNameLen-2] = '.';
	elem->name[maxNameLen-3] = '.';
      } else
	strcpy(elem->name, res);
  }

  return(NULL); /* NOTREACHED */
}

#endif /* defined(MULTITHREADED) && defined(ASYNC_ADDRESS_RESOLUTION) */

#endif

/* ************************************ */

void flowsProcess(const struct pcap_pkthdr *h, const u_char *p) {
  FlowFilterList *list = flowsList;
  
  while(list != NULL) {
    if(bpf_filter(list->filterCode, (u_char*)p, h->len, h->caplen)) {
      list->bytes += h->len;
      list->packets++;
    }
    
    list = list->next;
  }
}

/* ************************************ */

/*
 * This is the top level routine of the printer.  'p' is the points
 * to the ether header of the packet, 'tvp' is the timestamp,
 * 'length' is the length of the packet off the wire, and 'caplen'
 * is the number of bytes actually captured.
 */
void pbuf_process(u_char *unused, 
		  const struct pcap_pkthdr *h, 
		  const u_char *p)
{
  struct ether_header *ep;
  struct tokenRing_header *trp;
  u_int caplen = h->caplen;
  u_int length = h->len;
  const u_char *orig_p = p;
  u_char *ether_src, *ether_dst;
  unsigned short eth_type=0;
  /* Token-Ring Strings */
  struct tokenRing_llc *trllc;
  
  if(flowsList != NULL) /* Handle flows first */
    flowsProcess(h, p);

#ifdef MULTITHREADED
  accessMutex(&hostsHashMutex);
#endif

  memcpy(&lastPktTime, &h->ts, sizeof(lastPktTime));
 
  /* ethernet assumed */    
  if (caplen >= hlen) {
    HostTraffic *srcHost, *dstHost;
    
    ep = (struct ether_header *)p;

    switch(datalink) {
      case DLT_NULL: /* loopaback interface */
	/* 
	   Support for ethernet headerless interfaces (e.g. lo0)
	   Courtesy of Martin Kammerhofer <dada@sbox.tu-graz.ac.at> 
	*/
	length -= NULL_HDRLEN; /* don't count nullhdr */

	/* All this crap is due to the old little/big endian story... */
	if((p[0] == 0) && (p[1] == 0) && (p[2] == 8) && (p[3] == 0)) {
	  eth_type = ETHERTYPE_IP;
	  ether_src = ether_dst = dummyEthAddress;
	}
	break;
    case DLT_IEEE802: /* Token Ring */ 
      trp = (struct tokenRing_header*)p;
      ether_src = (u_char*)trp->trn_shost, ether_dst = (u_char*)trp->trn_dhost;

      hlen = sizeof(struct tokenRing_header) - 18;
      
      if (trp->trn_shost[0] & TR_RII) /* Source Routed Packet */
	hlen += ((ntohs(trp->trn_rcf) & TR_RCF_LEN_MASK) >> 8);
      
      length -= hlen, caplen -= hlen;
      
      p += hlen;
      trllc = (struct tokenRing_llc *)p;

      if (trllc->dsap == 0xAA && trllc->ssap == 0xAA)
	hlen = sizeof(struct tokenRing_llc);
      else
	hlen = sizeof(struct tokenRing_llc) - 5;
            
      length -= hlen, caplen -= hlen;
      
      p += hlen;	
	
      if (hlen == sizeof(struct tokenRing_llc))
	eth_type = ntohs(trllc->ethType);
      else
	eth_type = 0;
      break;
    default:
      eth_type = ntohs(ep->ether_type);      
      ether_src = ESRC(ep), ether_dst = EDST(ep);
    }    

    if((datalink == DLT_IEEE802) && (eth_type < ETHERMTU)) {
      trp = (struct tokenRing_header*)orig_p;
      ether_src = (u_char*)trp->trn_shost, ether_dst = (u_char*)trp->trn_dhost;
      srcHost = getHostInfo(NULL, ether_src);
      dstHost = getHostInfo(NULL, ether_dst);
      srcHost->otherSent += length;
      dstHost->otherReceived += length;	
    } if((datalink != DLT_IEEE802) && (eth_type <= ETHERMTU) && (length > 3)) {
      /* The code below has been taken from tcpdump */
      u_char *p1, sap_type;
      struct llc llcHeader;
      
      srcHost = getHostInfo(NULL, ether_src);
      dstHost = getHostInfo(NULL, ether_dst);
      
      p1 = (u_char*)(p+hlen);

      /* Watch out for possible alignment problems */
      memcpy(&llcHeader, (char*)p1, min(length, sizeof(llcHeader)));

      sap_type = llcHeader.ssap & ~LLC_GSAP;
      llcsap_string(sap_type);

      if ((llcHeader.ssap == LLCSAP_GLOBAL && llcHeader.dsap == LLCSAP_GLOBAL)
	  || (sap_type == 0xe0)) {
	/* IPX */
	srcHost->ipxSent += length;
	dstHost->ipxReceived += length;		
	ipxBytes += length;
      } else if (llcHeader.ssap == LLCSAP_NETBIOS 
		 && llcHeader.dsap == LLCSAP_NETBIOS) {
	/* Netbios */
	srcHost->netbiosSent += length;
	dstHost->netbiosReceived += length;	
	netbiosBytes += length;
      } else if ((sap_type == 0xf0) || (sap_type == 0xb4)
		 || (sap_type == 0xc4) || (sap_type == 0xf8)) {
	/* DLC (protocol used for printers) */
	srcHost->dlcSent += length;
	dstHost->dlcReceived += length;	
	dlcBytes += length;	
      } else {
	/* Unknown */

	/*
	  printf("[%u] [%x] %s %s > %s\n", (u_short)sap_type,(u_short)sap_type,
	  etheraddr_string(ether_src),
	  llcsap_string(llcHeader.ssap & ~LLC_GSAP),
	  etheraddr_string(ether_dst));
	*/

	srcHost->otherSent += length;
	dstHost->otherReceived += length;	
	otherBytes += length;
      }

      updatePacketCount(srcHost, dstHost, length);
    } else if(eth_type == ETHERTYPE_IP) {      
      if((datalink == DLT_IEEE802) && (eth_type > ETHERMTU))
	processIpPkt(p, length, ether_src, ether_dst);
      else
	processIpPkt(p+hlen, length, ether_src, ether_dst);
    } else {
      /* Non IP */
      srcHost = getHostInfo(NULL, ether_src);
      dstHost = getHostInfo(NULL, ether_dst);
      length -= hlen;

      updatePacketCount(srcHost, dstHost, length);
      
      switch(eth_type) 
	{	  
	case ETHERTYPE_ARP: /* ARP - Address resolution Protocol */
	case ETHERTYPE_REVARP: /* Reverse ARP */
	  srcHost->arp_rarpSent += length;
	  dstHost->arp_rarpReceived += length;
	  arpRarpBytes += length;
	  break;	  
	case ETHERTYPE_DN: /* Decnet */
	  srcHost->decnetSent += length;
	  dstHost->decnetReceived += length;
	  decnetBytes += length;
	  break;	  
	case ETHERTYPE_ATALK: /* AppleTalk */
	case ETHERTYPE_AARP:
	  srcHost->appletalkSent += length;
	  dstHost->appletalkReceived += length;
	  atalkBytes += length;
	  break;
	default:
	  srcHost->otherSent += length;
	  dstHost->otherReceived += length;
	  otherBytes += length;
	  break;
	}
    }
  }

  ethernetPkts++;
  ethernetBytes += length;

#ifdef MULTITHREADED
  if((actTime-lastThptUpdate) > (refreshRate*CLOCKS_PER_SEC)) {
    if(!webMode) 
      printHostsTraffic(0, 0);
  }
#endif

#ifdef MULTITHREADED
  releaseMutex(&hostsHashMutex);
#endif
}

